@@ -95,7 +95,6 @@ static int bs_sg_rw(int host_no, struct scsi_cmd *cmd)
 	unsigned char key = ILLEGAL_REQUEST;
 	uint16_t asc = ASC_LUN_NOT_SUPPORTED;
 
-	cmd->scsi_cmd_done = target_cmd_io_done;
 	ret = cmd->dev->bst->bs_cmd_submit(cmd);
 	if (ret) {
 		key = HARDWARE_ERROR;
@@ -251,8 +250,16 @@ static void bs_bsg_cmd_complete(int fd, int events, void *data)
 	 * Check SCSI: command completion status
 	 * */
 	if (!io_hdr.device_status) {
-		scsi_set_out_resid(cmd, io_hdr.dout_resid);
-		scsi_set_in_resid(cmd, io_hdr.din_resid);
+		uint32_t actual_len;
+
+		if (io_hdr.dout_resid) {
+			actual_len = scsi_get_out_length(cmd) - io_hdr.dout_resid;
+			scsi_set_out_resid_by_actual(cmd, actual_len);
+		}
+		if (io_hdr.din_resid) {
+			actual_len = scsi_get_in_length(cmd) - io_hdr.din_resid;
+			scsi_set_in_resid_by_actual(cmd, actual_len);
+		}
 	} else {
 		/*
 		 * NAB: Used by linux/block/bsg.c:bsg_ioctl(), is this
@@ -262,8 +269,8 @@ static void bs_bsg_cmd_complete(int fd, int events, void *data)
 		scsi_set_out_resid_by_actual(cmd, 0);
 		scsi_set_in_resid_by_actual(cmd, 0);
 	}
-	dprintf("[%d] BSG() Calling cmd->scsi_cmd_done()\n", getpid());
-	cmd->scsi_cmd_done(cmd, io_hdr.device_status);
+
+	target_cmd_io_done(cmd, io_hdr.device_status);
 }
 
 static void bs_sg_cmd_complete(int fd, int events, void *data)
@@ -271,7 +278,7 @@ static void bs_sg_cmd_complete(int fd, int events, void *data)
 	struct sg_io_hdr io_hdr;
 	struct scsi_cmd *cmd;
 	int err;
-	unsigned resid;
+	uint32_t actual_len;
 
 	memset(&io_hdr, 0, sizeof(io_hdr));
 	io_hdr.interface_id = 'S';
@@ -283,39 +290,50 @@ static void bs_sg_cmd_complete(int fd, int events, void *data)
 
 	cmd = (struct scsi_cmd *)io_hdr.usr_ptr;
 	if (!io_hdr.status) {
-		scsi_set_out_resid(cmd, io_hdr.resid);
-		scsi_set_in_resid(cmd, io_hdr.resid);
+		actual_len = io_hdr.dxfer_len - io_hdr.resid;
 	} else {
 		/* NO SENSE | ILI (Incorrect Length Indicator) */
 		if (io_hdr.sbp[2] == 0x20)
-			resid = io_hdr.dxfer_len - io_hdr.resid;
+			actual_len = io_hdr.dxfer_len - io_hdr.resid;
 		else
-			resid = 0;
+			actual_len = 0;
 
 		cmd->sense_len = io_hdr.sb_len_wr;
-		scsi_set_out_resid_by_actual(cmd, resid);
-		scsi_set_in_resid_by_actual(cmd, resid);
+	}
+	if (!actual_len || io_hdr.resid) {
+		if (io_hdr.dxfer_direction == SG_DXFER_TO_DEV)
+			scsi_set_out_resid_by_actual(cmd, actual_len);
+		else
+			scsi_set_in_resid_by_actual(cmd, actual_len);
 	}
 
-	cmd->scsi_cmd_done(cmd, io_hdr.status);
+	target_cmd_io_done(cmd, io_hdr.status);
 }
 
 static int get_bsg_major(char *path)
 {
 	FILE *devfd;
-	char majorno[8];
+	int majorno, n;
 	char dev[64];
 	char tmp[16];
 
 	sscanf(path, "/dev/bsg/%s", tmp);
 	sprintf(dev, "/sys/class/bsg/%s/dev", tmp);
 	devfd = fopen(dev, "r");
-	if (!devfd)
+	if (!devfd) {
+		eprintf("%s open failed errno: %d\n", dev, errno);
 		return -1;
-	fscanf(devfd, "%s:", majorno);
+	}
+	n = fscanf(devfd, "%d:", &majorno);
 	fclose(devfd);
-
-	return atoi(majorno);
+	if (n != 1) {
+		if (n < 0)
+			eprintf("reading major from %s failed errno: %d\n", dev, errno);
+		else
+			eprintf("reading major from %s failed: invalid input\n", dev);
+		return -1;
+	}
+	return majorno;
 }
 
 static int chk_sg_device(char *path)
@@ -402,7 +420,7 @@ static int init_sg_device(int fd)
 	return 0;
 }
 
-static int bs_sg_init(struct scsi_lu *lu)
+static tgtadm_err bs_sg_init(struct scsi_lu *lu)
 {
 	/*
 	 * Setup struct scsi_lu->cmd_perform() passthrough pointer
@@ -415,7 +433,7 @@ static int bs_sg_init(struct scsi_lu *lu)
 	 * usr/target.c:__cmd_done_passthrough().
 	 */
 	lu->cmd_done = &__cmd_done_passthrough;
-	return 0;
+	return TGTADM_SUCCESS;
 }
 
 static int bs_sg_open(struct scsi_lu *lu, char *path, int *fd, uint64_t *size)
@@ -463,17 +481,12 @@ static void bs_sg_close(struct scsi_lu *lu)
 	close(lu->fd);
 }
 
-static int bs_sg_cmd_done(struct scsi_cmd *cmd)
-{
-	return 0;
-}
-
-static int bs_sg_lu_init(struct scsi_lu *lu)
+static tgtadm_err bs_sg_lu_init(struct scsi_lu *lu)
 {
 	if (spc_lu_init(lu))
 		return TGTADM_NOMEM;
 
-	return 0;
+	return TGTADM_SUCCESS;
 }
 
 static struct backingstore_template sg_bst = {
@@ -483,7 +496,6 @@ static struct backingstore_template sg_bst = {
 	.bs_open		= bs_sg_open,
 	.bs_close		= bs_sg_close,
 	.bs_cmd_submit		= bs_sg_cmd_submit,
-	.bs_cmd_done		= bs_sg_cmd_done,
 };
 
 static struct backingstore_template bsg_bst = {
@@ -493,7 +505,6 @@ static struct backingstore_template bsg_bst = {
 	.bs_open		= bs_sg_open,
 	.bs_close		= bs_sg_close,
 	.bs_cmd_submit		= bs_bsg_cmd_submit,
-	.bs_cmd_done		= bs_sg_cmd_done,
 };
 
 static struct device_type_template sg_template = {
