@@ -113,7 +113,7 @@ static int resp_rewind(struct scsi_lu *lu)
 	return skip_next_header(lu);
 }
 
-static unsigned long current_size(struct scsi_cmd *cmd)
+static uint64_t current_size(struct scsi_cmd *cmd)
 {
 	struct ssc_info *ssc = dtype_priv(cmd->dev);
 	return ssc->c_blk.curr;
@@ -179,8 +179,8 @@ static int append_blk(struct scsi_cmd *cmd, uint8_t *data,
 
 	/* Write any data */
 	if (size) {
-		ret = pwrite(fd, data, size,
-			     (off_t)curr->curr + SSC_BLK_HDR_SIZE);
+		ret = pwrite64(fd, data, size,
+			       curr->curr + SSC_BLK_HDR_SIZE);
 		if (ret != size) {
 			eprintf("Write of data failed: %m\n");
 			sense_data_build(cmd, MEDIUM_ERROR, ASC_WRITE_ERROR);
@@ -189,8 +189,6 @@ static int append_blk(struct scsi_cmd *cmd, uint8_t *data,
 	}
 	/* Write new EOD blk header */
 
-	fsync(fd);
-
 	return SAM_STAT_GOOD;
 }
 
@@ -327,6 +325,9 @@ static int resp_var_read(struct scsi_cmd *cmd, uint8_t *buf, uint32_t length,
 		if (h->blk_type == BLK_EOD)
 			sense_data_build(cmd, 0x40 | BLANK_CHECK,
 					 NO_ADDITIONAL_SENSE);
+		else if (h->blk_type == BLK_FILEMARK)
+			ssc_sense_data_build(cmd, NO_SENSE | SENSE_FILEMARK,
+					     ASC_MARK, info, sizeof(info));
 		else
 			ssc_sense_data_build(cmd, NO_SENSE | 0x20,
 					     NO_ADDITIONAL_SENSE,
@@ -341,11 +342,14 @@ static int resp_var_read(struct scsi_cmd *cmd, uint8_t *buf, uint32_t length,
 
 		result = SAM_STAT_CHECK_CONDITION;
 
-		if (!length)
+		if (!length) {
+			if (h->blk_type == BLK_FILEMARK)
+				goto skip_and_out;
 			goto out;
+		}
 	}
 
-	ret = pread(cmd->dev->fd, buf, length, h->curr + SSC_BLK_HDR_SIZE);
+	ret = pread64(cmd->dev->fd, buf, length, h->curr + SSC_BLK_HDR_SIZE);
 	if (ret != length) {
 		sense_data_build(cmd, MEDIUM_ERROR, ASC_READ_ERROR);
 		result = SAM_STAT_CHECK_CONDITION;
@@ -353,6 +357,7 @@ static int resp_var_read(struct scsi_cmd *cmd, uint8_t *buf, uint32_t length,
 	}
 	*transferred = length;
 
+skip_and_out:
 	ret = skip_next_header(cmd->dev);
 	if (ret) {
 		sense_data_build(cmd, MEDIUM_ERROR, ASC_MEDIUM_FORMAT_CORRUPT);
@@ -399,8 +404,8 @@ static int resp_fixed_read(struct scsi_cmd *cmd, uint8_t *buf, uint32_t length,
 			goto out;
 		}
 
-		residue = pread(fd, buf, block_length,
-				h->curr + SSC_BLK_HDR_SIZE);
+		residue = pread64(fd, buf, block_length,
+				  h->curr + SSC_BLK_HDR_SIZE);
 		if (block_length != residue) {
 			eprintf("Could only read %d bytes, not %d\n",
 					(int)residue, block_length);
@@ -465,6 +470,7 @@ static void tape_rdwr_request(struct scsi_cmd *cmd)
 			append_blk(cmd, scsi_get_out_buffer(cmd), 0,
 					0, BLK_FILEMARK);
 
+		fsync(cmd->dev->fd);
 		break;
 
 	case READ_6:
@@ -569,7 +575,7 @@ static void tape_rdwr_request(struct scsi_cmd *cmd)
 		int len = scsi_get_in_length(cmd);
 
 		dprintf("Size of in_buffer = %d\n", len);
-		dprintf("Sizeof(buf): %Zd\n", sizeof(buf));
+		dprintf("Sizeof(buf): %zd\n", sizeof(buf));
 		dprintf("service action: 0x%02x\n", service_action);
 
 		if (service_action == 0) {	/* Short form - block ID */
@@ -604,7 +610,7 @@ static void tape_rdwr_request(struct scsi_cmd *cmd)
 			cmd, cmd->scb[0], ret, length, cmd->offset);
 }
 
-static int bs_ssc_init(struct scsi_lu *lu)
+static tgtadm_err bs_ssc_init(struct scsi_lu *lu)
 {
 	struct bs_thread_info *info = BS_THREAD_I(lu);
 	return bs_thread_open(info, tape_rdwr_request, 1);
@@ -620,7 +626,7 @@ static int bs_ssc_open(struct scsi_lu *lu, char *path, int *fd, uint64_t *size)
 
 	ssc = dtype_priv(lu);
 
-	*fd = backed_file_open(path, O_RDWR | O_LARGEFILE, size);
+	*fd = backed_file_open(path, O_RDWR | O_LARGEFILE, size, NULL);
 	if (*fd < 0) {
 		eprintf("Could not open %s %m\n", path);
 		return *fd;
@@ -686,11 +692,6 @@ static void bs_ssc_close(struct scsi_lu *lu)
 	close(lu->fd);
 }
 
-static int bs_ssc_cmd_done(struct scsi_cmd *cmd)
-{
-	return 0;
-}
-
 static struct backingstore_template ssc_bst = {
 	.bs_name		= "ssc",
 	.bs_datasize		= sizeof(struct bs_thread_info),
@@ -699,7 +700,6 @@ static struct backingstore_template ssc_bst = {
 	.bs_open		= bs_ssc_open,
 	.bs_close		= bs_ssc_close,
 	.bs_cmd_submit		= bs_thread_cmd_submit,
-	.bs_cmd_done		= bs_ssc_cmd_done,
 };
 
 __attribute__((constructor)) static void bs_ssc_constructor(void)
