@@ -76,6 +76,8 @@ void conn_exit(struct iscsi_connection *conn)
 	free(conn->req_buffer);
 	free(conn->rsp_buffer);
 	free(conn->initiator);
+	if (conn->initiator_alias)
+		free(conn->initiator_alias);
 
 	if (session)
 		session_put(session);
@@ -216,7 +218,7 @@ int conn_take_fd(struct iscsi_connection *conn)
 }
 
 /* called by tgtadm */
-int conn_close_admin(uint32_t tid, uint64_t sid, uint32_t cid)
+tgtadm_err conn_close_admin(uint32_t tid, uint64_t sid, uint32_t cid)
 {
 	struct iscsi_target* target = NULL;
 	struct iscsi_session *session;
@@ -242,3 +244,29 @@ int conn_close_admin(uint32_t tid, uint64_t sid, uint32_t cid)
 
 	return sess_found ? TGTADM_NO_CONNECTION : TGTADM_NO_SESSION;
 }
+
+void iscsi_update_conn_stats_rx(struct iscsi_connection *conn, int size, int opcode)
+{
+	conn->stats.rxdata_octets += (uint64_t)size;
+
+	if (unlikely(opcode < 0))
+		return;
+
+	if (opcode == ISCSI_OP_SCSI_CMD)
+		conn->stats.scsicmd_pdus++;
+	else if (opcode == ISCSI_OP_SCSI_DATA_OUT)
+		conn->stats.dataout_pdus++;
+}
+
+void iscsi_update_conn_stats_tx(struct iscsi_connection *conn, int size, int opcode)
+{
+	conn->stats.txdata_octets += (uint64_t)size;
+
+	if (unlikely(opcode < 0))
+		return;
+
+	if (opcode == ISCSI_OP_SCSI_DATA_IN)
+		conn->stats.datain_pdus++;
+	else if (opcode == ISCSI_OP_SCSI_CMD_RSP)
+		conn->stats.scsirsp_pdus++;
+}
