@@ -43,6 +43,9 @@
 
 #define MAX_QUEUE_CMD	128
 
+int default_nop_interval;
+int default_nop_count;
+
 LIST_HEAD(iscsi_portals_list);
 
 char *portal_arguments;
@@ -341,13 +344,16 @@ static void text_scan_login(struct iscsi_connection *conn)
 			}
 
 			err = param_check_val(session_keys, idx, &val);
-			if (idx == ISCSI_PARAM_MAX_XMIT_DLENGTH &&
-			    conn->session_type == SESSION_DISCOVERY)
+			if (err) {
+				text_key_add_reject(conn, key);
+				continue;
+			}
+			if (idx == ISCSI_PARAM_MAX_XMIT_DLENGTH)
 				conn->session_param[idx].val = val;
 			else
-				err = param_set_val(session_keys,
-						    conn->session_param,
-						    idx, &val);
+				param_set_val(session_keys,
+					      conn->session_param,
+					      idx, &val);
 
 			switch (conn->session_param[idx].state) {
 			case KEY_STATE_START:
@@ -453,7 +459,11 @@ static void login_start(struct iscsi_connection *conn)
 		return;
 	}
 	conn->initiator = strdup(name);
+
 	alias = text_key_find(conn, "InitiatorAlias");
+	if (alias)
+		conn->initiator_alias = strdup(alias);
+
 	session_type = text_key_find(conn, "SessionType");
 	target_name = text_key_find(conn, "TargetName");
 
@@ -529,6 +539,13 @@ static void login_start(struct iscsi_connection *conn)
 			return;
 		}
 
+		if (iqn_acl(conn->tid, conn)) {
+			rsp->status_class = ISCSI_STATUS_CLS_INITIATOR_ERR;
+			rsp->status_detail = ISCSI_LOGIN_STATUS_TGT_NOT_FOUND;
+			conn->state = STATE_EXIT;
+			return;
+		}
+
 		if (isns_scn_access(conn->tid, name)) {
 			rsp->status_class = ISCSI_STATUS_CLS_INITIATOR_ERR;
 			rsp->status_detail = ISCSI_LOGIN_STATUS_TGT_NOT_FOUND;
@@ -857,6 +874,10 @@ static void text_scan_text(struct iscsi_connection *conn)
 				continue;
 			}
 
+			/* strip zone id */
+			if (ss.ss_family == AF_INET6)
+				(void) strsep(&p, "%");
+
 			p = buf + strlen(buf);
 
 			if (ss.ss_family == AF_INET6)
@@ -994,36 +1015,63 @@ static void cmnd_finish(struct iscsi_connection *conn)
 	}
 }
 
-static void calc_residual(struct iscsi_cmd_rsp *rsp, struct iscsi_task *task)
+void iscsi_set_data_rsp_residual(struct iscsi_data_rsp *data_in, struct scsi_cmd *scmd)
 {
-	uint32_t residual = 0;
-	struct scsi_cmd *scmd = &task->scmd;
-	uint32_t read_len = scsi_get_in_length(scmd);
+	int32_t resid = scsi_get_in_resid(scmd);
 
-	/* we never have write under/over flow, no way to signal that
-	 * back from the target currently. */
-	if (scsi_get_data_dir(scmd) == DATA_BIDIRECTIONAL) {
-		if (task->len < read_len) {
-			rsp->flags |= ISCSI_FLAG_CMD_BIDI_UNDERFLOW;
-			residual = read_len - task->len;
-		} else if (task->len > read_len) {
-			rsp->flags |= ISCSI_FLAG_CMD_BIDI_OVERFLOW;
-			residual = task->len - read_len;
-		}
-		rsp->bi_residual_count = cpu_to_be32(residual);
-		rsp->residual_count = 0;
+	if (likely(!resid))
+		data_in->residual_count = 0;
+	else if (resid > 0) {
+		data_in->flags |= ISCSI_FLAG_CMD_UNDERFLOW;
+		data_in->residual_count = cpu_to_be32((uint32_t)resid);
 	} else {
-		if (task->len < read_len) {
-			rsp->flags |= ISCSI_FLAG_CMD_UNDERFLOW;
-			residual = read_len - task->len;
-		} else if (task->len > read_len) {
-			rsp->flags |= ISCSI_FLAG_CMD_OVERFLOW;
-			residual = task->len - read_len;
-		}
-		rsp->residual_count = cpu_to_be32(residual);
+		data_in->flags |= ISCSI_FLAG_CMD_OVERFLOW;
+		data_in->residual_count = cpu_to_be32((uint32_t)-resid);
 	}
 }
 
+static inline void iscsi_rsp_set_resid(struct iscsi_cmd_rsp *rsp,
+				       int32_t resid)
+{
+	if (likely(!resid))
+		rsp->residual_count = 0;
+	else if (resid > 0) {
+		rsp->flags |= ISCSI_FLAG_CMD_UNDERFLOW;
+		rsp->residual_count = cpu_to_be32((uint32_t)resid);
+	} else {
+		rsp->flags |= ISCSI_FLAG_CMD_OVERFLOW;
+		rsp->residual_count = cpu_to_be32((uint32_t)-resid);
+	}
+}
+
+static inline void iscsi_rsp_set_bidir_resid(struct iscsi_cmd_rsp *rsp,
+					     int32_t resid)
+{
+	if (likely(!resid))
+		rsp->bi_residual_count = 0;
+	else if (resid > 0) {
+		rsp->flags |= ISCSI_FLAG_CMD_BIDI_UNDERFLOW;
+		rsp->bi_residual_count = cpu_to_be32((uint32_t)resid);
+	} else {
+		rsp->flags |= ISCSI_FLAG_CMD_BIDI_OVERFLOW;
+		rsp->bi_residual_count = cpu_to_be32((uint32_t)-resid);
+	}
+}
+
+void iscsi_rsp_set_residual(struct iscsi_cmd_rsp *rsp, struct scsi_cmd *scmd)
+{
+	rsp->bi_residual_count = 0;
+	if (scsi_get_data_dir(scmd) == DATA_READ)
+		iscsi_rsp_set_resid(rsp, scsi_get_in_resid(scmd));
+	else if (scsi_get_data_dir(scmd) == DATA_WRITE)
+		iscsi_rsp_set_resid(rsp, scsi_get_out_resid(scmd));
+	else if (scsi_get_data_dir(scmd) == DATA_BIDIRECTIONAL) {
+		iscsi_rsp_set_bidir_resid(rsp, scsi_get_in_resid(scmd));
+		iscsi_rsp_set_resid(rsp, scsi_get_out_resid(scmd));
+	} else
+		rsp->residual_count = 0;
+}
+
 struct iscsi_sense_data {
 	uint16_t length;
 	uint8_t  data[0];
@@ -1046,7 +1094,7 @@ static int iscsi_cmd_rsp_build(struct iscsi_task *task)
 	rsp->exp_cmdsn = cpu_to_be32(conn->session->exp_cmd_sn);
 	rsp->max_cmdsn = cpu_to_be32(conn->session->exp_cmd_sn + MAX_QUEUE_CMD);
 
-	calc_residual(rsp, task);
+	iscsi_rsp_set_residual(rsp, &task->scmd);
 
 	sense_len = task->scmd.sense_len;
 	if (sense_len) {
@@ -1078,15 +1126,14 @@ static int iscsi_data_rsp_build(struct iscsi_task *task)
 	rsp->offset = cpu_to_be32(task->offset);
 	rsp->datasn = cpu_to_be32(task->exp_r2tsn++);
 
-	datalen = min_t(uint32_t, scsi_get_in_length(&task->scmd), task->len);
-	datalen -= task->offset;
+	datalen = scsi_get_in_transfer_len(&task->scmd) - task->offset;
 
 	maxdatalen = conn->tp->rdma ?
 		conn->session_param[ISCSI_PARAM_MAX_BURST].val :
 		conn->session_param[ISCSI_PARAM_MAX_XMIT_DLENGTH].val;
 
 	dprintf("%d %d %d %" PRIu32 "%x\n", datalen,
-		scsi_get_in_length(&task->scmd), task->len, maxdatalen,
+		scsi_get_in_transfer_len(&task->scmd), task->offset, maxdatalen,
 		rsp->itt);
 
 	if (datalen <= maxdatalen) {
@@ -1099,7 +1146,7 @@ static int iscsi_data_rsp_build(struct iscsi_task *task)
 			rsp->flags |= ISCSI_FLAG_DATA_STATUS;
 			rsp->cmd_status = result;
 			rsp->statsn = cpu_to_be32(conn->stat_sn++);
-			calc_residual((struct iscsi_cmd_rsp *) rsp, task);
+			iscsi_set_data_rsp_residual(rsp, &task->scmd);
 		}
 	} else
 		datalen = maxdatalen;
@@ -1200,7 +1247,6 @@ void iscsi_free_cmd_task(struct iscsi_task *task)
 static int iscsi_scsi_cmd_done(uint64_t nid, int result, struct scsi_cmd *scmd)
 {
 	struct iscsi_task *task = ITASK(scmd);
-	uint32_t read_len = scsi_get_in_length(scmd);
 
 	/*
 	 * Since the connection is closed we just free the task.
@@ -1213,19 +1259,6 @@ static int iscsi_scsi_cmd_done(uint64_t nid, int result, struct scsi_cmd *scmd)
 		return 0;
 	}
 
-	if (scsi_get_data_dir(scmd) == DATA_WRITE)
-		task->len = scsi_get_out_length(scmd) - scsi_get_out_resid(scmd);
-	else
-		task->len = scsi_get_in_length(scmd) - scsi_get_in_resid(scmd);
-
-	if (scsi_get_data_dir(scmd) == DATA_WRITE)
-		task->len = 0;  /* no read result */
-	else if (task->len > read_len) {
-		dprintf("shrunk too big device read len %d > %u\n",
-			task->len, read_len);
-		task->len = read_len;
-	}
-
 	list_add_tail(&task->c_list, &task->conn->tx_clist);
 	task->conn->tp->ep_event_modify(task->conn, EPOLLIN | EPOLLOUT);
 
@@ -1406,10 +1439,10 @@ static int iscsi_tm_execute(struct iscsi_task *task)
 		fn = ABORT_TASK_SET;
 		break;
 	case ISCSI_TM_FUNC_CLEAR_ACA:
-		fn = CLEAR_TASK_SET;
+		err = ISCSI_TMF_RSP_NOT_SUPPORTED;
 		break;
 	case ISCSI_TM_FUNC_CLEAR_TASK_SET:
-		fn = CLEAR_ACA;
+		err = ISCSI_TMF_RSP_NOT_SUPPORTED;
 		break;
 	case ISCSI_TM_FUNC_LOGICAL_UNIT_RESET:
 		fn = LOGICAL_UNIT_RESET;
@@ -1644,13 +1677,9 @@ static int iscsi_noop_out_rx_start(struct iscsi_connection *conn)
 
 	dprintf("%x %x %u\n", req->ttt, req->itt, ntoh24(req->dlength));
 	if (req->ttt != cpu_to_be32(ISCSI_RESERVED_TAG)) {
-		/*
-		 * We don't request a NOP-Out by sending a NOP-In.
-		 * See 10.18.2 in the draft 20.
-		 */
-		eprintf("initiator bug\n");
-		err = -ISCSI_REASON_PROTOCOL_ERROR;
-		goto out;
+		if ((req->opcode & ISCSI_OPCODE_MASK) == ISCSI_OP_NOOP_OUT) {
+			goto good;
+		}
 	}
 
 	if (req->itt == cpu_to_be32(ISCSI_RESERVED_TAG)) {
@@ -1661,6 +1690,7 @@ static int iscsi_noop_out_rx_start(struct iscsi_connection *conn)
 		}
 	}
 
+good:
 	conn->exp_stat_sn = be32_to_cpu(req->exp_statsn);
 
 	len = ntoh24(req->dlength);
@@ -1755,14 +1785,37 @@ static int iscsi_task_rx_start(struct iscsi_connection *conn)
 
 static int iscsi_scsi_cmd_tx_start(struct iscsi_task *task)
 {
+	enum data_direction data_dir = scsi_get_data_dir(&task->scmd);
 	int err = 0;
 
-	if (task->r2t_count)
-		err = iscsi_r2t_build(task);
-	else if (task->offset < task->len)
-		err = iscsi_data_rsp_build(task);
-	else
+	switch (data_dir) {
+	case DATA_NONE:
 		err = iscsi_cmd_rsp_build(task);
+		break;
+	case DATA_READ:
+		if (task->offset < scsi_get_in_transfer_len(&task->scmd))
+			err = iscsi_data_rsp_build(task);
+		else
+			err = iscsi_cmd_rsp_build(task);
+		break;
+	case DATA_WRITE:
+		if (task->r2t_count)
+			err = iscsi_r2t_build(task);
+		else
+			err = iscsi_cmd_rsp_build(task);
+		break;
+	case DATA_BIDIRECTIONAL:
+		if (task->r2t_count)
+			err = iscsi_r2t_build(task);
+		else if (task->offset < scsi_get_in_transfer_len(&task->scmd))
+			err = iscsi_data_rsp_build(task);
+		else
+			err = iscsi_cmd_rsp_build(task);
+		break;
+	default:
+		eprintf("Unexpected data_dir %d task %p\n", data_dir, task);
+		exit(-1);
+	}
 
 	return err;
 }
@@ -1783,6 +1836,28 @@ static int iscsi_logout_tx_start(struct iscsi_task *task)
 	return 0;
 }
 
+static int iscsi_noop_in_tx_start(struct iscsi_task *task)
+{
+	struct iscsi_connection *conn = task->conn;
+	struct iscsi_data_rsp *rsp = (struct iscsi_data_rsp *) &conn->rsp.bhs;
+
+	memset(rsp, 0, sizeof(*rsp));
+	rsp->opcode = ISCSI_OP_NOOP_IN;
+	rsp->flags = ISCSI_FLAG_CMD_FINAL;
+	rsp->itt = task->req.itt;
+	rsp->ttt = task->req.ttt;
+	rsp->statsn = cpu_to_be32(conn->stat_sn);
+	rsp->exp_cmdsn = cpu_to_be32(conn->session->exp_cmd_sn);
+	rsp->max_cmdsn = cpu_to_be32(conn->session->exp_cmd_sn + MAX_QUEUE_CMD);
+
+	/* TODO: honor max_burst */
+	conn->rsp.datasize = task->len;
+	hton24(rsp->dlength, task->len);
+	conn->rsp.data = task->data;
+
+	return 0;
+}
+
 static int iscsi_noop_out_tx_start(struct iscsi_task *task, int *is_rsp)
 {
 	struct iscsi_connection *conn = task->conn;
@@ -1790,6 +1865,10 @@ static int iscsi_noop_out_tx_start(struct iscsi_task *task, int *is_rsp)
 
 	if (task->req.itt == cpu_to_be32(ISCSI_RESERVED_TAG)) {
 		*is_rsp = 0;
+
+		if (conn->tp->ep_nop_reply)
+			conn->tp->ep_nop_reply(be32_to_cpu(task->req.ttt));
+
 		iscsi_free_task(task);
 	} else {
 		*is_rsp = 1;
@@ -1839,10 +1918,9 @@ static int iscsi_scsi_cmd_tx_done(struct iscsi_connection *conn)
 	case ISCSI_OP_R2T:
 		break;
 	case ISCSI_OP_SCSI_DATA_IN:
-		if (task->offset < task->len ||
+		if (task->offset < scsi_get_in_transfer_len(&task->scmd) ||
 		    scsi_get_result(&task->scmd) != SAM_STAT_GOOD ||
-		    scsi_get_data_dir(&task->scmd) == DATA_BIDIRECTIONAL ||
-		    conn->tp->rdma) {
+		    scsi_get_data_dir(&task->scmd) == DATA_BIDIRECTIONAL) {
 			dprintf("more data or sense or bidir %x\n", hdr->itt);
 			list_add(&task->c_list, &task->conn->tx_clist);
 			return 0;
@@ -1903,6 +1981,9 @@ static int iscsi_task_tx_start(struct iscsi_connection *conn)
 	case ISCSI_OP_SCSI_CMD:
 		err = iscsi_scsi_cmd_tx_start(task);
 		break;
+	case ISCSI_OP_NOOP_IN:
+		err = iscsi_noop_in_tx_start(task);
+		break;
 	case ISCSI_OP_NOOP_OUT:
 		err = iscsi_noop_out_tx_start(task, &is_rsp);
 		if (!is_rsp)
@@ -1927,7 +2008,7 @@ nodata:
 
 static int do_recv(struct iscsi_connection *conn, int next_state)
 {
-	int ret;
+	int ret, opcode;
 
 	ret = conn->tp->ep_read(conn, conn->rx_buffer, conn->rx_size);
 	if (!ret) {
@@ -1940,20 +2021,12 @@ static int do_recv(struct iscsi_connection *conn, int next_state)
 			return -EIO;
 	}
 
-	conn->stats.rxdata_octets += ret;
 	conn->rx_size -= ret;
 	conn->rx_buffer += ret;
 
-	if (conn->rx_iostate == IOSTATE_RX_BHS) {
-		switch (conn->req.bhs.opcode & ISCSI_OPCODE_MASK) {
-		case ISCSI_OP_SCSI_CMD:
-			conn->stats.scsicmd_pdus++;
-			break;
-		case ISCSI_OP_SCSI_DATA_OUT:
-			conn->stats.dataout_pdus++;
-		}
-	}
-
+	opcode = (conn->rx_iostate == IOSTATE_RX_BHS) ?
+		(conn->req.bhs.opcode & ISCSI_OPCODE_MASK) : -1;
+	iscsi_update_conn_stats_rx(conn, ret, opcode);
 
 	if (!conn->rx_size)
 		conn->rx_iostate = next_state;
@@ -2125,7 +2198,7 @@ again:
 
 static int do_send(struct iscsi_connection *conn, int next_state)
 {
-	int ret;
+	int ret, opcode;
 again:
 	ret = conn->tp->ep_write_begin(conn, conn->tx_buffer, conn->tx_size);
 	if (ret < 0) {
@@ -2137,20 +2210,12 @@ again:
 		return -EIO;
 	}
 
-	conn->stats.txdata_octets += ret;
 	conn->tx_size -= ret;
 	conn->tx_buffer += ret;
 
-	if (conn->tx_iostate == IOSTATE_TX_BHS) {
-		switch (conn->rsp.bhs.opcode) {
-		case ISCSI_OP_SCSI_DATA_IN:
-			conn->stats.datain_pdus++;
-			break;
-		case ISCSI_OP_SCSI_CMD_RSP:
-			conn->stats.scsirsp_pdus++;
-			break;
-		}
-	}
+	opcode = (conn->tx_iostate == IOSTATE_TX_BHS) ?
+			(conn->req.bhs.opcode & ISCSI_OPCODE_MASK) : -1;
+	iscsi_update_conn_stats_tx(conn, ret, opcode);
 
 	if (conn->tx_size)
 		goto again;
@@ -2414,6 +2479,10 @@ int iscsi_param_parse_portals(char *p, int do_add,
 					return -1;
 				}
 			}
+		} else if (!strncmp(p, "nop_interval", 12)) {
+			iscsi_set_nop_interval(atoi(p+13));
+		} else if (!strncmp(p, "nop_count", 9)) {
+			iscsi_set_nop_count(atoi(p+10));
 		}
 
 		p += strcspn(p, ",");
@@ -2442,7 +2511,6 @@ static int iscsi_portal_destroy(char *p)
 
 static struct tgt_driver iscsi = {
 	.name			= "iscsi",
-	.use_kernel		= 0,
 	.init			= iscsi_init,
 	.exit			= iscsi_exit,
 	.target_create		= iscsi_target_create,
@@ -2453,6 +2521,7 @@ static struct tgt_driver iscsi = {
 
 	.update			= iscsi_target_update,
 	.show			= iscsi_target_show,
+	.stat			= iscsi_stat,
 	.cmd_end_notify		= iscsi_scsi_cmd_done,
 	.mgmt_end_notify	= iscsi_tm_done,
 	.transportid		= iscsi_transportid,
