@@ -28,6 +28,7 @@
 #include "param.h"
 #include "log.h"
 #include "tgtd.h"
+#include "util.h"
 
 #include "iscsi_proto.h"
 #include "iscsi_if.h"
@@ -73,6 +74,7 @@ struct iscsi_session {
 	struct list_head hlist;
 
 	char *initiator;
+	char *initiator_alias;
 	struct iscsi_target *target;
 	uint8_t isid[6];
 	uint16_t tsih;
@@ -149,6 +151,7 @@ struct iscsi_connection {
 	struct param session_param[ISCSI_PARAM_MAX];
 
 	char *initiator;
+	char *initiator_alias;
 	uint8_t isid[6];
 	uint16_t tsih;
 	uint16_t cid;
@@ -236,6 +239,9 @@ struct iscsi_connection {
 
 #define INCOMING_BUFSIZE	8192
 
+extern int default_nop_interval;
+extern int default_nop_count;
+
 struct iscsi_target {
 	struct list_head tlist;
 
@@ -259,6 +265,8 @@ struct iscsi_target {
 	struct list_head isns_list;
 
 	int rdma;
+	int nop_interval;
+	int nop_count;
 };
 
 enum task_flags {
@@ -301,7 +309,7 @@ extern int conn_get(struct iscsi_connection *conn);
 extern struct iscsi_connection * conn_find(struct iscsi_session *session, uint32_t cid);
 extern int conn_take_fd(struct iscsi_connection *conn);
 extern void conn_add_to_session(struct iscsi_connection *conn, struct iscsi_session *session);
-extern int conn_close_admin(uint32_t tid, uint64_t sid, uint32_t cid);
+extern tgtadm_err conn_close_admin(uint32_t tid, uint64_t sid, uint32_t cid);
 
 /* iscsid.c */
 extern char *text_key_find(struct iscsi_connection *conn, char *searchKey);
@@ -312,8 +320,16 @@ extern void iscsi_rx_handler(struct iscsi_connection *conn);
 extern int iscsi_scsi_cmd_execute(struct iscsi_task *task);
 extern int iscsi_transportid(int tid, uint64_t itn_id, char *buf, int size);
 extern int iscsi_add_portal(char *addr, int port, int tpgt);
+extern void iscsi_print_nop_settings(struct concat_buf *b, int tid);
+extern int iscsi_update_target_nop_count(int tid, int count);
+extern int iscsi_update_target_nop_interval(int tid, int interval);
+extern void iscsi_set_nop_interval(int interval);
+extern void iscsi_set_nop_count(int count);
 extern int iscsi_delete_portal(char *addr, int port);
 extern int iscsi_param_parse_portals(char *p, int do_add, int do_delete);
+extern void iscsi_update_conn_stats_rx(struct iscsi_connection *conn, int size, int opcode);
+extern void iscsi_update_conn_stats_tx(struct iscsi_connection *conn, int size, int opcode);
+extern void iscsi_rsp_set_residual(struct iscsi_cmd_rsp *rsp, struct scsi_cmd *scmd);
 
 /* iscsid.c iscsi_task */
 extern void iscsi_free_task(struct iscsi_task *task);
@@ -327,21 +343,24 @@ extern void session_get(struct iscsi_session *session);
 extern void session_put(struct iscsi_session *session);
 
 /* target.c */
-struct iscsi_target * target_find_by_name(const char *name);
-struct iscsi_target * target_find_by_id(int tid);
+extern struct iscsi_target * target_find_by_name(const char *name);
+extern struct iscsi_target * target_find_by_id(int tid);
 extern void target_list_build(struct iscsi_connection *, char *, char *);
 extern int ip_acl(int tid, struct iscsi_connection *conn);
+extern int iqn_acl(int tid, struct iscsi_connection *conn);
 extern int iscsi_target_create(struct target *);
-extern void iscsi_target_destroy(int);
-extern int iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid,
-			     uint64_t lun, char *buf, int rest);
-int iscsi_target_update(int mode, int op, int tid, uint64_t sid, uint64_t lun,
-			uint32_t cid, char *name);
-int target_redirected(struct iscsi_target *target,
-	struct iscsi_connection *conn, char *buf, int *reason);
+extern void iscsi_target_destroy(int tid, int force);
+extern tgtadm_err iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid,
+				    uint64_t lun, struct concat_buf *b);
+extern tgtadm_err iscsi_stat(int mode, int tid, uint64_t sid, uint32_t cid,
+			     uint64_t lun, struct concat_buf *b);
+extern tgtadm_err iscsi_target_update(int mode, int op, int tid, uint64_t sid, uint64_t lun,
+				      uint32_t cid, char *name);
+extern int target_redirected(struct iscsi_target *target,
+			     struct iscsi_connection *conn, char *buf, int *reason);
 
 /* param.c */
-int param_index_by_name(char *name, struct iscsi_key *keys);
+extern int param_index_by_name(char *name, struct iscsi_key *keys);
 
 /* transport.c */
 extern int iscsi_init(int, char *);
@@ -350,19 +369,10 @@ extern void iscsi_exit(void);
 /* isns.c */
 extern int isns_init(void);
 extern void isns_exit(void);
-extern int isns_show(char *buf, int rest);
-extern int isns_update(char *params);
+extern tgtadm_err isns_show(struct concat_buf *b);
+extern tgtadm_err isns_update(char *params);
 extern int isns_scn_access(int tid, char *name);
 extern int isns_target_register(char *name);
 extern int isns_target_deregister(char *name);
 
-#define buffer_check(buf, total, len, rest)	\
-({						\
-	buf += len;				\
-	total += len;				\
-	rest -= len;				\
-	if (!rest)				\
-		break;				\
-})
-
 #endif	/* ISCSID_H */
