@@ -32,6 +32,8 @@
 #include <sys/epoll.h>
 #include <infiniband/verbs.h>
 #include <rdma/rdma_cma.h>
+#include <sys/ipc.h>
+#include <sys/shm.h>
 
 #include "util.h"
 #include "iscsid.h"
@@ -80,7 +82,9 @@ char *iser_portal_addr;
 
 #define MAX_POLL_WC 32
 
+#define DEFAULT_POOL_SIZE_MB    1024
 #define ISER_MAX_QUEUE_CMD      128     /* iSCSI cmd window size */
+#define MAX_CQ_ENTRIES          (128 * 1024)
 
 #define MASK_BY_BIT(b)  	((1UL << b) - 1)
 #define ALIGN_TO_BIT(x, b)      ((((unsigned long)x) + MASK_BY_BIT(b)) & \
@@ -92,13 +96,10 @@ struct iscsi_sense_data {
 	uint8_t data[0];
 } __packed;
 
-/*
- * Number of allocatable data buffers, each of this size.  Do at least 128
- * for linux iser.  The membuf size is rounded up at initialization time
- * to the hardware page size so that allocations for direct IO devices are
- * aligned.
- */
-static int membuf_num = 16 * ISER_MAX_QUEUE_CMD;
+static size_t buf_pool_sz_mb = DEFAULT_POOL_SIZE_MB;
+static int cq_vector = -1;
+
+static int membuf_num;
 static size_t membuf_size = RDMA_TRANSFER_SIZE;
 
 static int iser_conn_get(struct iser_conn *conn);
@@ -134,7 +135,7 @@ static inline void schedule_task_iosubmit(struct iser_task *task,
 	list_add_tail(&task->exec_list, &conn->iosubmit_list);
 	tgt_add_sched_event(&conn->sched_iosubmit);
 
-	dprintf("task:%p tag:0x%04"PRIx64 "cmdsn:0x%x\n",
+	dprintf("task:%p tag:0x%04"PRIx64 " cmdsn:0x%x\n",
 		task, task->tag, task->cmd_sn);
 }
 
@@ -144,7 +145,7 @@ static inline void schedule_rdma_read(struct iser_task *task,
 	list_add_tail(&task->rdma_list, &conn->rdma_rd_list);
 	tgt_add_sched_event(&conn->sched_rdma_rd);
 
-	dprintf("task:%p tag:0x%04"PRIx64 "cmdsn:0x%x\n",
+	dprintf("task:%p tag:0x%04"PRIx64 " cmdsn:0x%x\n",
 		task, task->tag, task->cmd_sn);
 }
 
@@ -154,7 +155,7 @@ static inline void schedule_resp_tx(struct iser_task *task,
 	list_add_tail(&task->tx_list, &conn->resp_tx_list);
 	tgt_add_sched_event(&conn->sched_tx);
 
-	dprintf("task:%p tag:0x%04"PRIx64 "cmdsn:0x%x\n",
+	dprintf("task:%p tag:0x%04"PRIx64 " cmdsn:0x%x\n",
 		task, task->tag, task->cmd_sn);
 }
 
@@ -164,7 +165,7 @@ static inline void schedule_post_recv(struct iser_task *task,
 	list_add_tail(&task->recv_list, &conn->post_recv_list);
 	tgt_add_sched_event(&conn->sched_post_recv);
 
-	dprintf("task:%p tag:0x%04"PRIx64 "cmdsn:0x%x\n",
+	dprintf("task:%p tag:0x%04"PRIx64 " cmdsn:0x%x\n",
 		task, task->tag, task->cmd_sn);
 }
 
@@ -393,8 +394,8 @@ static void iser_prep_resp_send_req(struct iser_task *task,
 	txd->send_wr.next = (next_wr ? &next_wr->send_wr : NULL);
 	txd->send_wr.send_flags = (signaled ? IBV_SEND_SIGNALED : 0);
 
-	dprintf("task:%p wr_id:0x%"PRIx64 "tag:0x%04"PRIx64 ""
-		"dtbuf:%p dtsz:%u ahs_sz:%u stat:0x%x statsn:0x%x expcmdsn:0x%x\n",
+	dprintf("task:%p wr_id:0x%"PRIx64 " tag:0x%04"PRIx64 " dtbuf:%p "
+		"dtsz:%u ahs_sz:%u stat:0x%x statsn:0x%x expcmdsn:0x%x\n",
 		task, txd->send_wr.wr_id, task->tag,
 		pdu->membuf.addr, pdu->membuf.size, pdu->ahssize,
 		bhs->rsvd2[1], ntohl(bhs->statsn), ntohl(bhs->exp_statsn));
@@ -431,8 +432,8 @@ static void iser_prep_rdma_wr_send_req(struct iser_task *task,
 	/* offset += rdmad->sge.length // ToDo: multiple RDMA-Write buffers */
 	rdmad->send_wr.wr.rdma.rkey = task->rem_read_stag;
 
-	dprintf(" task:%p tag:0x%04"PRIx64 "wr_id:0x%"PRIx64 "daddr:0x%"PRIx64 "dsz:%u "
-		"bufsz:%u rdma:%d lkey:%x raddr:%"PRIx64 "rkey:%x rems:%u\n",
+	dprintf(" task:%p tag:0x%04"PRIx64 " wr_id:0x%"PRIx64 " daddr:0x%"PRIx64 " dsz:%u "
+		"bufsz:%u rdma:%d lkey:%x raddr:%"PRIx64 " rkey:%x rems:%u\n",
 		task, task->tag, rdmad->send_wr.wr_id, rdmad->sge.addr,
 		rdmad->sge.length, rdma_buf->size, rdma_buf->rdma,
 		rdmad->sge.lkey, rdmad->send_wr.wr.rdma.remote_addr,
@@ -475,8 +476,8 @@ static void iser_prep_rdma_rd_send_req(struct iser_task *task,
 	rdmad->send_wr.wr.rdma.rkey =
 		(uint32_t) task->rem_write_stag;
 
-	dprintf("task:%p wr_id:0x%"PRIx64 "tag:0x%04"PRIx64 "daddr:0x%"PRIx64 "dsz:%u "
-		"bufsz:%u rdma:%d lkey:%x raddr:%"PRIx64 "rkey:%x rems:%u\n",
+	dprintf("task:%p wr_id:0x%"PRIx64 " tag:0x%04"PRIx64 " daddr:0x%"PRIx64 " dsz:%u "
+		"bufsz:%u rdma:%d lkey:%x raddr:%"PRIx64 " rkey:%x rems:%u\n",
 		task, rdmad->send_wr.wr_id, task->tag, rdmad->sge.addr,
 		rdmad->sge.length, rdma_buf->size, rdma_buf->rdma,
 		rdmad->sge.lkey, rdmad->send_wr.wr.rdma.remote_addr,
@@ -559,16 +560,70 @@ static inline void iser_set_rsp_stat_sn(struct iscsi_session *session,
 	}
 }
 
+static uint8_t* iser_alloc_pool(size_t pool_size, int *shmid)
+{
+	int shmemid;
+	uint8_t *buf;
+
+	/* allocate memory */
+	shmemid = shmget(IPC_PRIVATE, pool_size,
+			SHM_HUGETLB | IPC_CREAT | SHM_R | SHM_W);
+
+	if (shmemid < 0) {
+		eprintf("shmget rdma pool sz:%zu failed\n", pool_size);
+		goto failed_huge_page;
+	}
+
+	/* get pointer to allocated memory */
+	buf = shmat(shmemid, NULL, 0);
+
+	if (buf == (void*)-1) {
+		eprintf("Shared memory attach failure (errno=%d %m)", errno);
+		shmctl(shmemid, IPC_RMID, NULL);
+		goto failed_huge_page;
+	}
+
+	/* mark 'to be destroyed' when process detaches from shmem segment
+	   this will clear the HugePage resources even if process if killed not nicely.
+	   From checking shmctl man page it is unlikely that it will fail here. */
+	if (shmctl(shmemid, IPC_RMID, NULL)) {
+		eprintf("Shared memory contrl mark 'to be destroyed' failed (errno=%d %m)", errno);
+	}
+
+	dprintf("Allocated huge page sz:%zu\n", pool_size);
+	*shmid = shmemid;
+	return buf;
+
+ failed_huge_page:
+	*shmid = -1;
+	return valloc(pool_size);
+}
+
+static void iser_free_pool(uint8_t *pool_buf, int shmid) {
+	if (shmid >= 0) {
+		if (shmdt(pool_buf) != 0) {
+			eprintf("shmem detach failure (errno=%d %m)", errno);
+		}
+	} else {
+		free(pool_buf);
+	}
+}
+
 static int iser_init_rdma_buf_pool(struct iser_device *dev)
 {
 	uint8_t *pool_buf, *list_buf;
 	size_t pool_size, list_size;
 	struct iser_membuf *rdma_buf;
+	int shmid;
 	int i;
 
+	/* The membuf size is rounded up at initialization time to the hardware
+	   page size so that allocations for direct IO devices are aligned. */
 	membuf_size = roundup(membuf_size, pagesize);
-	pool_size = membuf_num * membuf_size;
-	pool_buf = valloc(pool_size);
+	pool_size = buf_pool_sz_mb * 1024 * 1024;
+	membuf_num = pool_size / membuf_size;
+	pool_size = membuf_num * membuf_size; /* reflect possible round-down */
+	pool_buf = iser_alloc_pool(pool_size, &shmid);
 	if (!pool_buf) {
 		eprintf("malloc rdma pool sz:%zu failed\n", pool_size);
 		return -ENOMEM;
@@ -578,7 +633,7 @@ static int iser_init_rdma_buf_pool(struct iser_device *dev)
 	list_buf = malloc(list_size);
 	if (!list_buf) {
 		eprintf("malloc list_buf sz:%zu failed\n", list_size);
-		free(pool_buf);
+		iser_free_pool(pool_buf, shmid);
 		return -ENOMEM;
 	}
 
@@ -587,13 +642,14 @@ static int iser_init_rdma_buf_pool(struct iser_device *dev)
 				    IBV_ACCESS_LOCAL_WRITE);
 	if (!dev->membuf_mr) {
 		eprintf("ibv_reg_mr failed, %m\n");
-		free(pool_buf);
+		iser_free_pool(pool_buf, shmid);
 		free(list_buf);
 		return -1;
 	}
 	dprintf("pool buf:%p list:%p mr:%p lkey:0x%x\n",
 		pool_buf, list_buf, dev->membuf_mr, dev->membuf_mr->lkey);
 
+	dev->rdma_hugetbl_shmid = shmid;
 	dev->membuf_regbuf = pool_buf;
 	dev->membuf_listbuf = list_buf;
 	INIT_LIST_HEAD(&dev->membuf_free);
@@ -615,6 +671,24 @@ static int iser_init_rdma_buf_pool(struct iser_device *dev)
 	return 0;
 }
 
+static void iser_destroy_rdma_buf_pool(struct iser_device *dev)
+{
+        int err;
+
+	assert(list_empty(&dev->membuf_alloc));
+
+	err = ibv_dereg_mr(dev->membuf_mr);
+	if (err)
+		eprintf("ibv_dereg_mr failed: (errno=%d %m)\n", errno);
+
+	iser_free_pool(dev->membuf_regbuf, dev->rdma_hugetbl_shmid);
+	free(dev->membuf_listbuf);
+
+	dev->membuf_mr = NULL;
+	dev->membuf_regbuf = NULL;
+	dev->membuf_listbuf = NULL;
+}
+
 static struct iser_membuf *iser_dev_alloc_rdma_buf(struct iser_device *dev)
 {
 	struct iser_membuf *rdma_buf;
@@ -1044,14 +1118,16 @@ static void iser_free_ff_resources(struct iser_conn *conn)
 int iser_login_complete(struct iscsi_connection *iscsi_conn)
 {
 	struct iser_conn *conn = ISER_CONN(iscsi_conn);
-	unsigned int irdsl, trdsl, outst_pdu, hdrsz;
+	unsigned int trdsl;
+	/* unsigned int irdsl; */
+	unsigned int outst_pdu, hdrsz;
 	int err = -1;
 
 	dprintf("entry\n");
 
 	/* one more send, then done; login resources are left until then */
 	iser_conn_login_phase_set(conn, LOGIN_PHASE_LAST_SEND);
-	irdsl = iscsi_conn->session_param[ISCSI_PARAM_INITIATOR_RDSL].val;
+	/* irdsl = iscsi_conn->session_param[ISCSI_PARAM_INITIATOR_RDSL].val; */
 	trdsl = iscsi_conn->session_param[ISCSI_PARAM_TARGET_RDSL].val;
 
 	/* ToDo: outstanding pdus num */
@@ -1676,7 +1752,7 @@ static int iser_logout_exec(struct iser_task *task)
 
 	/* add the logout resp to tx list, and force all scheduled tx now */
 	list_add_tail(&task->tx_list, &conn->resp_tx_list);
-	dprintf("add to tx list, logout resp task:%p tag:0x%04"PRIx64 "cmdsn:0x%x\n",
+	dprintf("add to tx list, logout resp task:%p tag:0x%04"PRIx64 " cmdsn:0x%x\n",
 		task, task->tag, task->cmd_sn);
 	tgt_remove_sched_event(&conn->sched_tx);
 	iser_sched_tx(&conn->sched_tx);
@@ -1910,7 +1986,7 @@ static void iser_task_free_dout_tasks(struct iser_task *task)
 	}
 }
 
-static void iser_scsi_cmd_iosubmit(struct iser_task *task)
+static void iser_scsi_cmd_iosubmit(struct iser_task *task, int not_last)
 {
 	struct iscsi_cmd *req_bhs = (struct iscsi_cmd *) task->pdu.bhs;
 	struct scsi_cmd *scmd = &task->scmd;
@@ -1919,11 +1995,12 @@ static void iser_scsi_cmd_iosubmit(struct iser_task *task)
 	struct iser_membuf *data_buf;
 
 	scmd->state = 0;
+	if (not_last)
+		set_cmd_not_last(scmd);
 	scsi_set_in_length(scmd, 0);
 	scsi_set_out_length(scmd, 0);
 
 	if (task->is_write) {
-		scsi_set_out_resid(scmd, 0);
 		/* It's either the last buffer, which is RDMA,
 		   or the only buffer */
 		data_buf = list_entry(task->out_buf_list.prev,
@@ -1946,7 +2023,6 @@ static void iser_scsi_cmd_iosubmit(struct iser_task *task)
 		scsi_set_out_length(scmd, task->out_len);
 	}
 	if (task->is_read) {
-		scsi_set_in_resid(scmd, 0);
 		/* ToDo: multiple RDMA-Read buffers */
 		data_buf = list_entry(task->in_buf_list.next,
 				      struct iser_membuf, task_list);
@@ -1972,32 +2048,6 @@ static void iser_scsi_cmd_iosubmit(struct iser_task *task)
 	target_cmd_queue(session->target->tid, scmd);
 }
 
-static void iser_rsp_set_read_resid(struct iscsi_cmd_rsp *rsp_bhs,
-				    int in_resid)
-{
-	if (in_resid > 0) {
-		rsp_bhs->flags |= ISCSI_FLAG_CMD_UNDERFLOW;
-		rsp_bhs->residual_count = cpu_to_be32((uint32_t)in_resid);
-	} else {
-		rsp_bhs->flags |= ISCSI_FLAG_CMD_OVERFLOW;
-		rsp_bhs->residual_count = cpu_to_be32(((uint32_t)-in_resid));
-	}
-	rsp_bhs->bi_residual_count = 0;
-}
-
-static void iser_rsp_set_bidir_resid(struct iscsi_cmd_rsp *rsp_bhs,
-				     int in_resid)
-{
-	if (in_resid > 0) {
-		rsp_bhs->flags |= ISCSI_FLAG_CMD_BIDI_UNDERFLOW;
-		rsp_bhs->bi_residual_count = cpu_to_be32((uint32_t)in_resid);
-	} else {
-		rsp_bhs->flags |= ISCSI_FLAG_CMD_BIDI_OVERFLOW;
-		rsp_bhs->bi_residual_count = cpu_to_be32(((uint32_t)-in_resid));
-	}
-	rsp_bhs->residual_count = 0;
-}
-
 static int iser_scsi_cmd_done(uint64_t nid, int result,
 			      struct scsi_cmd *scmd)
 {
@@ -2007,7 +2057,6 @@ static int iser_scsi_cmd_done(uint64_t nid, int result,
 	struct iser_conn *conn = task->conn;
 	struct iscsi_session *session = conn->h.session;
 	unsigned char sense_len = scmd->sense_len;
-	int in_resid;
 
 	assert(nid == scmd->cmd_itn_id);
 
@@ -2031,21 +2080,12 @@ static int iser_scsi_cmd_done(uint64_t nid, int result,
 	iser_set_rsp_stat_sn(session, task->pdu.bhs);
 	rsp_bhs->exp_datasn = 0;
 
+	iscsi_rsp_set_residual(rsp_bhs, scmd);
 	if (task->is_read) {
-		in_resid = scsi_get_in_resid(scmd);
-		if (in_resid != 0) {
-			if (!task->is_write)
-				iser_rsp_set_read_resid(rsp_bhs, in_resid);
-			else
-				iser_rsp_set_bidir_resid(rsp_bhs, in_resid);
-			if (in_resid > 0)
-				task->rdma_wr_remains = task->in_len - in_resid;
-		}
-		/* schedule_rdma_write(task, conn); // ToDo: need this? */
-	} else {
-		rsp_bhs->bi_residual_count = 0;
-		rsp_bhs->residual_count = 0;
+		task->rdma_wr_remains = scsi_get_in_transfer_len(scmd);
+		task->rdma_wr_sz = scsi_get_in_transfer_len(scmd);
 	}
+
 	task->pdu.ahssize = 0;
 	task->pdu.membuf.size = 0;
 
@@ -2060,7 +2100,7 @@ static int iser_scsi_cmd_done(uint64_t nid, int result,
 		memcpy(sense->data, scmd->sense_buffer, sense_len);
 		task->pdu.membuf.size = sense_len + sizeof(*sense);
 	}
-	dprintf("task:%p tag:0x%04"PRIx64 "status:%x statsn:%d flags:0x%x "
+	dprintf("task:%p tag:0x%04"PRIx64 " status:%x statsn:%d flags:0x%x "
 		"in_len:%d out_len:%d rsd:%d birsd:%d\n",
 		task, task->tag, rsp_bhs->cmd_status,
 		conn->h.stat_sn-1, rsp_bhs->flags,
@@ -2094,7 +2134,7 @@ static void iser_sched_buf_alloc(struct event_data *evt)
 	}
 }
 
-static inline int task_batch_type(struct iser_task *task)
+static inline int task_can_batch(struct iser_task *task)
 {
 	struct iscsi_cmd *req_bhs = (struct iscsi_cmd *) task->pdu.bhs;
 
@@ -2118,13 +2158,17 @@ static inline int task_batch_type(struct iser_task *task)
 static void iser_sched_iosubmit(struct event_data *evt)
 {
 	struct iser_conn *conn = (struct iser_conn *) evt->data;
-	struct iser_task *task = NULL;
+	struct iser_task *task, *next_task;
+	int last_in_batch;
 
-	while (!list_empty(&conn->iosubmit_list)) {
-		task = list_first_entry(&conn->iosubmit_list,
-					struct iser_task, exec_list);
+	list_for_each_entry_safe(task, next_task, &conn->iosubmit_list, exec_list) {
+		if (&next_task->exec_list == &conn->iosubmit_list)
+			last_in_batch = 1; /* end of list */
+		else
+			last_in_batch = (task_can_batch(task) &&
+					 task_can_batch(next_task)) ? 0 : 1;
 		list_del(&task->exec_list);
-		iser_scsi_cmd_iosubmit(task);
+		iser_scsi_cmd_iosubmit(task, !last_in_batch);
 	}
 }
 
@@ -2343,7 +2387,6 @@ static int iser_scsi_cmd_rx(struct iser_task *task)
 			task->unsol_remains = 0;
 			task->rdma_rd_sz = 0;
 			task->rdma_rd_remains = 0;
-
 		} else {
 			scsi_set_data_dir(&task->scmd, DATA_NONE);
 			task->out_len = 0;
@@ -2395,7 +2438,7 @@ static int iser_scsi_cmd_rx(struct iser_task *task)
 
 	list_add_tail(&task->session_list, &session->cmd_list);
 out:
-	dprintf("task:%p tag:0x%04"PRIx64 "scsi_op:0x%x %s%s in_len:%d out_len:%d "
+	dprintf("task:%p tag:0x%04"PRIx64 " scsi_op:0x%x %s%s in_len:%d out_len:%d "
 		"imm_sz:%d unsol_sz:%d cmdsn:0x%x expcmdsn:0x%x\n",
 		task, task->tag, req_bhs->cdb[0],
 		task->is_read ? "rd" : "", task->is_write ? "wr" : "",
@@ -2579,7 +2622,7 @@ static int iser_queue_task(struct iscsi_session *session,
 	int err;
 
 	if (unlikely(task->is_immediate)) {
-		dprintf("exec imm task task:%p tag:0x%0" PRIx64 "cmd_sn:0x%x\n",
+		dprintf("exec imm task task:%p tag:0x%0"PRIx64 " cmd_sn:0x%x\n",
 			task, task->tag, cmd_sn);
 		err = iser_task_delivery(task);
 		if (likely(!err || err == -ENOMEM))
@@ -2654,6 +2697,8 @@ static int iser_parse_req_headers(struct iser_task *task)
 	struct iser_conn *conn = task->conn;
 	struct iser_hdr *iser_hdr = task->pdu.iser_hdr;
 	struct iscsi_hdr *iscsi_hdr = task->pdu.bhs;
+	unsigned pdu_dlength = ntoh24(iscsi_hdr->dlength);
+	unsigned pdu_len = pdu_dlength + sizeof(struct iscsi_hdr);
 	int err = -1;
 
 	switch (iser_hdr->flags & 0xF0) {
@@ -2691,12 +2736,14 @@ static int iser_parse_req_headers(struct iser_task *task)
 
 	task->pdu.ahssize = iscsi_hdr->hlength * 4;
 	task->pdu.membuf.addr += task->pdu.ahssize;
-	task->pdu.membuf.size = ntoh24(iscsi_hdr->dlength);
+	pdu_len += task->pdu.ahssize;
+	task->pdu.membuf.size = pdu_dlength;
 	task->pdu.membuf.rdma = 0;
 
 	task->tag = iscsi_hdr->itt;
 	task->cmd_sn = be32_to_cpu(iscsi_hdr->statsn);
 	conn->h.exp_stat_sn = be32_to_cpu(iscsi_hdr->exp_statsn);
+	iscsi_update_conn_stats_rx(&conn->h, pdu_len, task->opcode);
 
 	return err;
 }
@@ -2821,9 +2868,11 @@ static void iser_tx_complete_handler(struct iser_work_req *txd)
 {
 	struct iser_task *task = txd->task;
 	struct iser_conn *conn = task->conn;
+	int opcode = task->pdu.bhs->opcode & ISCSI_OPCODE_MASK;
 
-	dprintf("conn:%p task:%p tag:0x%04"PRIx64 "\n",
-		&conn->h, task, task->tag);
+	iscsi_update_conn_stats_tx(&conn->h, txd->sge.length, opcode);
+	dprintf("conn:%p task:%p tag:0x%04"PRIx64 " opcode:0x%x\n",
+		&conn->h, task, task->tag, opcode);
 	iser_conn_put(conn);
 
 	list_del(&task->tx_list); /* remove from conn->sent_list */
@@ -2846,6 +2895,7 @@ static void iser_rdma_wr_complete_handler(struct iser_work_req *rdmad)
 	struct iser_task *task = rdmad->task;
 	struct iser_conn *conn = task->conn;
 
+	iscsi_update_conn_stats_tx(&conn->h, rdmad->sge.length, ISCSI_OP_SCSI_DATA_IN);
 	dprintf("conn:%p task:%p tag:0x%04"PRIx64 "\n",
 		&conn->h, task, task->tag);
 	iser_conn_put(conn);
@@ -2859,6 +2909,7 @@ static void iser_rdma_rd_complete_handler(struct iser_work_req *rdmad)
 	struct iser_task *task = rdmad->task;
 	struct iser_conn *conn = task->conn;
 
+	iscsi_update_conn_stats_rx(&conn->h, rdmad->sge.length, ISCSI_OP_SCSI_DATA_OUT);
 	task->rdma_rd_remains -= rdmad->sge.length;
 	dprintf("conn:%p task:%p tag:0x%04"PRIx64 ", rems rdma:%d unsol:%d\n",
 		&conn->h, task, task->tag, task->rdma_rd_remains,
@@ -2918,13 +2969,13 @@ static void handle_wc_error(struct ibv_wc *wc)
 	struct iser_conn *conn = task ? task->conn : NULL;
 
 	if (wc->status != IBV_WC_WR_FLUSH_ERR)
-		eprintf("conn:%p task:%p tag:0x%04"PRIx64 "wr_id:0x%p op:%s "
+		eprintf("conn:%p task:%p tag:0x%04"PRIx64 " wr_id:0x%p op:%s "
 			"err:%s vendor_err:0x%0x\n",
 			&conn->h, task, task->tag, req,
 			iser_ib_op_to_str(req->iser_ib_op),
 			ibv_wc_status_str(wc->status), wc->vendor_err);
 	else
-		dprintf("conn:%p task:%p tag:0x%04"PRIx64 "wr_id:0x%p op:%s "
+		dprintf("conn:%p task:%p tag:0x%04"PRIx64 " wr_id:0x%p op:%s "
 			"err:%s vendor_err:0x%0x\n",
 			&conn->h, task, task->tag, req,
 			iser_ib_op_to_str(req->iser_ib_op),
@@ -3204,7 +3255,7 @@ static int iser_device_init(struct iser_device *dev)
 		eprintf("ibv_query_device failed, %m\n");
 		goto out;
 	}
-	cqe_num = dev->device_attr.max_cqe;
+	cqe_num = min(dev->device_attr.max_cqe, MAX_CQ_ENTRIES);
 	dprintf("max %d CQEs\n", cqe_num);
 
 	err = -1;
@@ -3214,8 +3265,18 @@ static int iser_device_init(struct iser_device *dev)
 		goto out;
 	}
 
+	/* verify cq_vector */
+	if (cq_vector < 0)
+		cq_vector = control_port % dev->ibv_ctxt->num_comp_vectors;
+	else if (cq_vector >= dev->ibv_ctxt->num_comp_vectors) {
+		eprintf("Bad CQ vector. max: %d\n",
+			dev->ibv_ctxt->num_comp_vectors);
+		goto out;
+	}
+	dprintf("CQ vector: %d\n", cq_vector);
+
 	dev->cq = ibv_create_cq(dev->ibv_ctxt, cqe_num, NULL,
-				dev->cq_channel, 0);
+				dev->cq_channel, cq_vector);
 	if (dev->cq == NULL) {
 		eprintf("ibv_create_cq failed\n");
 		goto out;
@@ -3249,6 +3310,31 @@ out:
 	return err;
 }
 
+static void iser_device_release(struct iser_device *dev)
+{
+	int err;
+
+	list_del(&dev->list);
+
+	tgt_event_del(dev->ibv_ctxt->async_fd);
+	tgt_event_del(dev->cq_channel->fd);
+	tgt_remove_sched_event(&dev->poll_sched);
+
+	err = ibv_destroy_cq(dev->cq);
+	if (err)
+		eprintf("ibv_destroy_cq failed: (errno=%d %m)\n", errno);
+
+	err = ibv_destroy_comp_channel(dev->cq_channel);
+	if (err)
+		eprintf("ibv_destroy_comp_channel failed: (errno=%d %m)\n", errno);
+
+	iser_destroy_rdma_buf_pool(dev);
+
+	err = ibv_dealloc_pd(dev->pd);
+	if (err)
+		eprintf("ibv_dealloc_pd failed: (errno=%d %m)\n", errno);
+}
+
 /*
  * Init entire iscsi transport.  Begin listening for connections.
  */
@@ -3308,6 +3394,27 @@ static int iser_ib_init(void)
 	return err;
 }
 
+static void iser_ib_release(void)
+{
+	int err;
+	struct iser_device *dev, *tdev;
+
+	assert(list_empty(&iser_conn_list));
+
+	list_for_each_entry_safe(dev, tdev, &iser_dev_list, list) {
+	        iser_device_release(dev);
+		free(dev);
+	}
+
+	tgt_event_del(cma_listen_id->channel->fd);
+
+	err = rdma_destroy_id(cma_listen_id);
+	if (err)
+		eprintf("rdma_destroy_id failed: (errno=%d %m)\n", errno);
+
+	rdma_destroy_event_channel(rdma_evt_channel);
+}
+
 static int iser_send_nop = 1;
 static struct tgt_work nop_work;
 
@@ -3336,8 +3443,10 @@ static int iser_init(int index, char *args)
 	int err;
 
 	err = iser_ib_init();
-	if (err)
+	if (err) {
+		iser_send_nop = 0;
 		return err;
+	}
 
 	if (iser_send_nop) {
 		nop_work.func = iser_nop_work_handler;
@@ -3353,6 +3462,8 @@ static void iser_exit(void)
 {
 	if (iser_send_nop)
 		del_work(&nop_work);
+
+	iser_ib_release();
 }
 
 static int iser_target_create(struct target *t)
@@ -3378,6 +3489,8 @@ static const char *lld_param_port = "port";
 static const char *lld_param_nop = "nop";
 static const char *lld_param_on = "on";
 static const char *lld_param_off = "off";
+static const char *lld_param_pool_sz_mb = "pool_sz_mb";
+static const char *lld_param_cq_vector = "cq_vector";
 
 static int iser_param_parser(char *p)
 {
@@ -3410,6 +3523,22 @@ static int iser_param_parser(char *p)
 				err = -EINVAL;
 				break;
 			}
+		} else if (!strncmp(p, lld_param_pool_sz_mb,
+				    strlen(lld_param_pool_sz_mb))) {
+			q = p + strlen(lld_param_pool_sz_mb) + 1;
+			buf_pool_sz_mb = atoi(q);
+			if (buf_pool_sz_mb < 128)
+				buf_pool_sz_mb = 128;
+		} else if (!strncmp(p, lld_param_cq_vector,
+				    strlen(lld_param_cq_vector))) {
+			q = p + strlen(lld_param_cq_vector) + 1;
+			cq_vector = atoi(q);
+			if (cq_vector < 0) {
+				eprintf("unsupported value for param: %s\n",
+					lld_param_cq_vector);
+				err = -EINVAL;
+				break;
+			}
 		} else {
 			dprintf("unsupported param:%s\n", p);
 			err = -EINVAL;
@@ -3435,7 +3564,6 @@ static struct iscsi_transport iscsi_iser = {
 
 static struct tgt_driver iser = {
 	.name   		= "iser",
-	.use_kernel     	= 0,
 	.init   		= iser_init,
 	.exit   		= iser_exit,
 	.target_create  	= iser_target_create,
@@ -3443,6 +3571,7 @@ static struct tgt_driver iser = {
 
 	.update 		= iscsi_target_update,
 	.show 			= iscsi_target_show,
+	.stat                   = iscsi_stat,
 	.cmd_end_notify 	= iser_scsi_cmd_done,
 	.mgmt_end_notify	= iser_tm_done,
 	.transportid    	= iscsi_transportid,
