@@ -284,7 +284,12 @@ static void iser_login_oper_scan(struct iscsi_connection *iscsi_conn,
 			}
 
 			err = param_check_val(session_keys, idx, &val);
-			err = param_set_val(session_keys, iscsi_conn->session_param, idx, &val);
+			if (err) {
+				iser_text_key_add_reject(iscsi_conn, tx_pdu, key);
+				continue;
+			}
+
+			param_set_val(session_keys, iscsi_conn->session_param, idx, &val);
 
 			switch (iscsi_conn->session_param[idx].state) {
 			case KEY_STATE_START:
@@ -434,7 +439,11 @@ static void iser_login_start(struct iscsi_connection *iscsi_conn,
 		return;
 	}
 	iscsi_conn->initiator = strdup(name);
+
 	alias = iser_text_key_find(req_data, req_datasize, "InitiatorAlias");
+	if (alias)
+		iscsi_conn->initiator_alias = strdup(alias);
+
 	session_type = iser_text_key_find(req_data, req_datasize, "SessionType");
 	target_name = iser_text_key_find(req_data, req_datasize, "TargetName");
 
@@ -499,6 +508,13 @@ static void iser_login_start(struct iscsi_connection *iscsi_conn,
 			return;
 		}
 
+		if (iqn_acl(iscsi_conn->tid, iscsi_conn)) {
+			rsp_bhs->status_class = ISCSI_STATUS_CLS_INITIATOR_ERR;
+			rsp_bhs->status_detail = ISCSI_LOGIN_STATUS_TGT_NOT_FOUND;
+			iscsi_conn->state = STATE_EXIT;
+			return;
+		}
+
 		if (isns_scn_access(iscsi_conn->tid, name)) {
 			rsp_bhs->status_class = ISCSI_STATUS_CLS_INITIATOR_ERR;
 			rsp_bhs->status_detail = ISCSI_LOGIN_STATUS_TGT_NOT_FOUND;
@@ -594,6 +610,8 @@ void iser_login_exec(struct iscsi_connection *iscsi_conn,
 	struct iscsi_login_rsp *rsp_bhs = (struct iscsi_login_rsp *)tx_pdu->bhs;
 	int stay = 0, nsg_disagree = 0;
 
+	tx_pdu->membuf.size = 0;
+
 	memset(rsp_bhs, 0, BHS_SIZE);
 	if ((req_bhs->opcode & ISCSI_OPCODE_MASK) != ISCSI_OP_LOGIN ||
 	    !(req_bhs->opcode & ISCSI_OP_IMMEDIATE)) {
@@ -847,4 +865,3 @@ int iser_text_exec(struct iscsi_connection *iscsi_conn,
 
 	return 0;
 }
-
