@@ -578,6 +578,8 @@ static int recv_hdr(int fd, struct isns_io *rx, struct isns_hdr *hdr)
 	flags = ntohs(hdr->flags);						\
 	transaction = ntohs(hdr->transaction);					\
 	sequence = ntohs(hdr->sequence);					\
+	dprintf("got a header %x %u %x %u %u\n", function, length, flags,	\
+		transaction, sequence);						\
 }
 
 static int recv_pdu(int fd, struct isns_io *rx, struct isns_hdr *hdr)
@@ -591,8 +593,6 @@ static int recv_pdu(int fd, struct isns_io *rx, struct isns_hdr *hdr)
 
 	/* Now we got a complete header */
 	get_hdr_param(hdr, function, length, flags, transaction, sequence);
-	dprintf("got a header %x %u %x %u %u\n", function, length, flags,
-		transaction, sequence);
 
 	if (length + sizeof(*hdr) > BUFSIZE) {
 		eprintf("FIXME we cannot handle this yet %u!\n", length);
@@ -774,7 +774,6 @@ static void isns_handle(int fd, int events, void *data)
 	int err;
 	struct isns_io *rx = &isns_rx;
 	struct isns_hdr *hdr = (struct isns_hdr *) rx->buf;
-	uint32_t result;
 	uint16_t function, length, flags, transaction, sequence;
 	char *name = NULL;
 
@@ -790,7 +789,6 @@ static void isns_handle(int fd, int events, void *data)
 	}
 
 	get_hdr_param(hdr, function, length, flags, transaction, sequence);
-	result = ntohl((uint32_t) hdr->pdu[0]);
 
 	switch (function) {
 	case ISNS_FUNC_DEV_ATTR_REG_RSP:
@@ -1068,20 +1066,16 @@ void isns_exit(void)
 	free(rxbuf);
 }
 
-int isns_show(char *buf, int rest)
+tgtadm_err isns_show(struct concat_buf *b)
 {
-	int total = 0, max = rest;
+	concat_printf(b, "iSNS:\n");
+	concat_printf(b, _TAB1 "iSNS=%s\n", use_isns ? "On" : "Off");
+	concat_printf(b, _TAB1 "iSNSServerIP=%s\n", isns_addr);
+	concat_printf(b, _TAB1 "iSNSServerPort=%d\n", isns_port);
+	concat_printf(b, _TAB1 "iSNSAccessControl=%s\n",
+		      use_isns_ac ? "On" : "Off");
 
-	shprintf(total, buf, rest, "iSNS:\n");
-	shprintf(total, buf, rest, _TAB1 "iSNS=%s\n",
-		 use_isns ? "On" : "Off");
-	shprintf(total, buf, rest, _TAB1 "iSNSServerIP=%s\n", isns_addr);
-	shprintf(total, buf, rest, _TAB1 "iSNSServerPort=%d\n", isns_port);
-	shprintf(total, buf, rest, _TAB1 "iSNSAccessControl=%s\n",
-		 use_isns_ac ? "On" : "Off");
-	return total;
-overflow:
-	return max;
+	return TGTADM_SUCCESS;
 }
 
 enum {
@@ -1097,9 +1091,9 @@ static match_table_t tokens = {
 	{Opt_err, NULL},
 };
 
-int isns_update(char *params)
+tgtadm_err isns_update(char *params)
 {
-	int ret = 0;
+    tgtadm_err adm_err = TGTADM_SUCCESS;
 	char *p;
 
 	while ((p = strsep(&params, ",")) != NULL) {
@@ -1119,7 +1113,7 @@ int isns_update(char *params)
 			else if (!strcmp(tmp, "Off"))
 				isns = 0;
 			else {
-				ret = TGTADM_INVALID_REQUEST;
+				adm_err = TGTADM_INVALID_REQUEST;
 				break;
 			}
 			if (use_isns == isns)
@@ -1134,7 +1128,7 @@ int isns_update(char *params)
 			break;
 		case Opt_port:
 			if (match_int(&args[0], &isns_port))
-				ret = TGTADM_INVALID_REQUEST;
+				adm_err = TGTADM_INVALID_REQUEST;
 			break;
 		case Opt_ac:
 			match_strncpy(tmp, &args[0], sizeof(tmp));
@@ -1142,12 +1136,12 @@ int isns_update(char *params)
 			break;
 		case Opt_state:
 			match_strncpy(tmp, &args[0], sizeof(tmp));
-			system_set_state(tmp);
+			adm_err = system_set_state(tmp);
 			break;
 		default:
-			ret = TGTADM_INVALID_REQUEST;
+			adm_err = TGTADM_INVALID_REQUEST;
 		}
 	}
 
-	return ret;
+	return adm_err;
 }
