@@ -36,6 +36,7 @@
 #include "tgtadm.h"
 #include "tgtd.h"
 #include "target.h"
+#include "util.h"
 
 LIST_HEAD(iscsi_targets_list);
 
@@ -188,6 +189,35 @@ int ip_acl(int tid, struct iscsi_connection *conn)
 	}
 	return -EPERM;
 }
+
+static int iqn_match(struct iscsi_connection *conn, char *name)
+{
+	return strcmp(conn->initiator, name);
+}
+
+int iqn_acl(int tid, struct iscsi_connection *conn)
+{
+	int idx, enable, err;
+	char *name;
+
+	enable = 0;
+	for (idx = 0;; idx++) {
+		name = iqn_acl_get(tid, idx);
+		if (!name)
+			break;
+
+		enable = 1;
+		err = iqn_match(conn, name);
+		if (!err)
+			return 0;
+	}
+
+	if (!enable)
+		return 0;
+	else
+		return -EPERM;
+}
+
 static int
 get_redirect_address(char *callback, char *buffer, int buflen,
 			char **address, char **ip_port, int *rsn)
@@ -325,6 +355,9 @@ void target_list_build(struct iscsi_connection *conn, char *addr, char *name)
 		if (ip_acl(target->tid, conn))
 			continue;
 
+		if (iqn_acl(target->tid, conn))
+			continue;
+
 		if (isns_scn_access(target->tid, conn->initiator))
 			continue;
 
@@ -357,9 +390,11 @@ struct iscsi_target* target_find_by_id(int tid)
 	return NULL;
 }
 
-void iscsi_target_destroy(int tid)
+void iscsi_target_destroy(int tid, int force)
 {
 	struct iscsi_target* target;
+	struct iscsi_session *session, *stmp;
+	struct iscsi_connection *conn, *ctmp;
 
 	target = target_find_by_id(tid);
 	if (!target) {
@@ -367,10 +402,17 @@ void iscsi_target_destroy(int tid)
 		return;
 	}
 
-	if (target->nr_sessions) {
+	if (!force && target->nr_sessions) {
 		eprintf("the target %d still has sessions\n", tid);
 		return;
 	}
+
+	list_for_each_entry_safe(session, stmp, &target->sessions_list, slist) {
+		list_for_each_entry_safe(conn, ctmp, &session->conn_list, clist) {
+			conn_close(conn);
+		}
+	}
+
 	if (!list_empty(&target->sessions_list)) {
 		eprintf("bug still have sessions %d\n", tid);
 		exit(-1);
@@ -428,6 +470,8 @@ int iscsi_target_create(struct target *t)
 	INIT_LIST_HEAD(&target->sessions_list);
 	INIT_LIST_HEAD(&target->isns_list);
 	target->tid = tid;
+	target->nop_interval = default_nop_interval;
+	target->nop_count = default_nop_count;
 	list_add_tail(&target->tlist, &iscsi_targets_list);
 
 	isns_target_register(tgt_targetname(tid));
@@ -454,16 +498,17 @@ static int iscsi_session_param_update(struct iscsi_target* target, int idx, char
 	return 0;
 }
 
-int iscsi_target_update(int mode, int op, int tid, uint64_t sid, uint64_t lun,
-			uint32_t cid, char *name)
+tgtadm_err iscsi_target_update(int mode, int op, int tid, uint64_t sid, uint64_t lun,
+			       uint32_t cid, char *name)
 {
-	int idx, err = TGTADM_INVALID_REQUEST;
+	tgtadm_err adm_err = TGTADM_INVALID_REQUEST;
+	int idx, err;
 	char *str;
 	struct iscsi_target* target;
 
 	switch (mode) {
 	case MODE_SYSTEM:
-		err = isns_update(name);
+		adm_err = isns_update(name);
 		break;
 	case MODE_TARGET:
 		target = target_find_by_id(tid);
@@ -477,74 +522,75 @@ int iscsi_target_update(int mode, int op, int tid, uint64_t sid, uint64_t lun,
 		if (!strncmp(name, "RedirectAddress", 15)) {
 			snprintf(target->redirect_info.addr,
 				 sizeof(target->redirect_info.addr), "%s", str);
-			err = TGTADM_SUCCESS;
+			adm_err = TGTADM_SUCCESS;
 			break;
 		} else if (!strncmp(name, "RedirectPort", 12)) {
 			snprintf(target->redirect_info.port,
 				 sizeof(target->redirect_info.port), "%s", str);
-			err = TGTADM_SUCCESS;
+			adm_err = TGTADM_SUCCESS;
 			break;
 		} else if (!strncmp(name, "RedirectReason", 14)) {
 			if (!strncmp(str, "Temporary", 9)) {
 				target->redirect_info.reason =
 					ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP;
-				err = TGTADM_SUCCESS;
+				adm_err = TGTADM_SUCCESS;
 			} else if (!strncmp(str, "Permanent", 9)) {
 				target->redirect_info.reason =
 					ISCSI_LOGIN_STATUS_TGT_MOVED_PERM;
-				err = TGTADM_SUCCESS;
+				adm_err = TGTADM_SUCCESS;
 			} else
 				break;
 		} else if (!strncmp(name, "RedirectCallback", 16)) {
 			target->redirect_info.callback = strdup(str);
 			if (!target->redirect_info.callback) {
-				err = TGTADM_NOMEM;
+				adm_err = TGTADM_NOMEM;
 				break;
 			}
-			err = TGTADM_SUCCESS;
+			adm_err = TGTADM_SUCCESS;
+		} else if (!strncmp(name, "nop_count", 9)) {
+			err = iscsi_update_target_nop_count(tid,
+				atoi(&name[10]));
+			adm_err = !err ? TGTADM_SUCCESS :
+				TGTADM_INVALID_REQUEST;
+			break;
+		} else if (!strncmp(name, "nop_interval", 12)) {
+			err = iscsi_update_target_nop_interval(tid,
+				atoi(&name[13]));
+			adm_err = !err ? TGTADM_SUCCESS :
+				TGTADM_INVALID_REQUEST;
+			break;
 		}
 
 		idx = param_index_by_name(name, session_keys);
 		if (idx >= 0) {
 			err = iscsi_session_param_update(target, idx, str);
-			if (err < 0)
-				err = TGTADM_INVALID_REQUEST;
+			adm_err = !err ? TGTADM_SUCCESS : TGTADM_INVALID_REQUEST;
 		}
 		break;
 	case MODE_CONNECTION:
 		if (op == OP_DELETE)
-			err = conn_close_admin(tid, sid, cid);
+			adm_err = conn_close_admin(tid, sid, cid);
 		break;
 	default:
 		break;
 	}
-	return err;
+	return adm_err;
 }
 
-static int show_iscsi_param(char *buf, struct param *param, int rest)
+static tgtadm_err show_iscsi_param(struct param *param, struct concat_buf *b)
 {
-	int i, len, total;
-	char value[64];
 	struct iscsi_key *keys = session_keys;
+	int i;
+	char value[64];
 
-	for (i = total = 0; session_keys[i].name; i++) {
+	for (i = 0; session_keys[i].name; i++) {
 		param_val_to_str(keys, i, param[i].val, value);
-		len = snprintf(buf, rest, "%s=%s\n", keys[i].name, value);
-		buffer_check(buf, total, len, rest);
+		concat_printf(b, "%s=%s\n", keys[i].name, value);
 	}
 
-	return total;
+	return TGTADM_SUCCESS;
 }
 
-#define __buffer_check(buf, total, len, rest)	\
-({						\
-	buf += len;				\
-	total += len;				\
-	rest -= len;				\
-	if (!rest)				\
-		return total;			\
-})
-
 static struct iscsi_session *iscsi_target_find_session(
 	struct iscsi_target *target,
 	uint64_t sid)
@@ -560,71 +606,34 @@ static struct iscsi_session *iscsi_target_find_session(
 
 }
 
-static int iscsi_target_show_session(struct iscsi_target *target, uint64_t sid,
-				     char *buf, int rest)
+static tgtadm_err iscsi_target_show_session(struct iscsi_target *target, uint64_t sid,
+					    struct concat_buf *b)
 {
-	int len = 0, total = 0;
+	tgtadm_err adm_err = TGTADM_SUCCESS;
 	struct iscsi_session *session;
 
 	session = iscsi_target_find_session(target, sid);
+	if (session)
+		adm_err = show_iscsi_param(session->session_param, b);
 
-	if (session) {
-		len = show_iscsi_param(buf, session->session_param, rest);
-		__buffer_check(buf, total, len, rest);
-
-	}
-
-	return total;
+	return adm_err;
 }
 
-static int iscsi_target_show_stats(struct iscsi_target *target, uint64_t sid,
-				   char *buf, int rest)
-{
-	int len = 0, total = 0;
-	struct iscsi_session *session;
-	struct iscsi_connection *conn;
-
-	session = iscsi_target_find_session(target, sid);
-
-	if (session) {
-		list_for_each_entry(conn, &session->conn_list, clist) {
-			len = snprintf(buf, rest,
-				       "rxdata_octets: %" PRIu64 "\n"
-				       "txdata_octets: %" PRIu64 "\n"
-				       "dataout_pdus:  %d\n"
-				       "datain_pdus:   %d\n"
-				       "scsicmd_pdus:  %d\n"
-				       "scsirsp_pdus:  %d\n",
-				       conn->stats.rxdata_octets,
-				       conn->stats.txdata_octets,
-				       conn->stats.dataout_pdus,
-				       conn->stats.datain_pdus,
-				       conn->stats.scsicmd_pdus,
-				       conn->stats.scsirsp_pdus);
-			__buffer_check(buf, total, len, rest);
-		}
-	}
-
-	return total;
-
-}
-
-static int iscsi_target_show_connections(struct iscsi_target *target,
-					 uint64_t sid,
-					 char *buf, int rest)
+static tgtadm_err iscsi_target_show_connections(struct iscsi_target *target,
+						uint64_t sid,
+						struct concat_buf *b)
 {
+	tgtadm_err adm_err = TGTADM_SUCCESS;
 	struct iscsi_session *session;
 	struct iscsi_connection *conn;
 	char addr[128];
-	int len = 0, total = 0;
 
 	list_for_each_entry(session, &target->sessions_list, slist) {
 		list_for_each_entry(conn, &session->conn_list, clist) {
 			memset(addr, 0, sizeof(addr));
 			conn->tp->ep_show(conn, addr, sizeof(addr));
 
-
-			len = snprintf(buf, rest, "Session: %u\n"
+			concat_printf(b, "Session: %u\n"
 				_TAB1 "Connection: %u\n"
 				_TAB2 "Initiator: %s\n"
 				_TAB2 "%s\n",
@@ -632,16 +641,15 @@ static int iscsi_target_show_connections(struct iscsi_target *target,
 				conn->cid,
 				session->initiator,
 				addr);
-			__buffer_check(buf, total, len, rest);
 		}
 	}
-	return total;
+	return adm_err;
 }
 
-static int iscsi_target_show_portals(struct iscsi_target *target, uint64_t sid,
-				   char *buf, int rest)
+static tgtadm_err iscsi_target_show_portals(struct iscsi_target *target, uint64_t sid,
+					    struct concat_buf *b)
 {
-	int len = 0, total = 0;
+	tgtadm_err adm_err = TGTADM_SUCCESS;
 	struct iscsi_portal *portal;
 
 	list_for_each_entry(portal, &iscsi_portals_list,
@@ -649,96 +657,212 @@ static int iscsi_target_show_portals(struct iscsi_target *target, uint64_t sid,
 		int is_ipv6;
 
 		is_ipv6 = strchr(portal->addr, ':') != NULL;
-		len = snprintf(buf, rest,
-			       "Portal: %s%s%s:%d,%d\n",
+		concat_printf(b, "Portal: %s%s%s:%d,%d\n",
 			       is_ipv6 ? "[" : "",
 			       portal->addr,
 			       is_ipv6 ? "]" : "",
 			       portal->port ? portal->port : ISCSI_LISTEN_PORT,
 			       portal->tpgt);
-		__buffer_check(buf, total, len, rest);
 	}
 
-	return total;
-
+	return adm_err;
 }
 
-static int show_redirect_info(struct iscsi_target* target, char *buf, int rest)
+static tgtadm_err show_redirect_info(struct iscsi_target *target, struct concat_buf *b)
 {
-	int len, total = 0;
+	tgtadm_err adm_err = TGTADM_SUCCESS;
 
-	len = snprintf(buf, rest, "RedirectAddress=%s\n", target->redirect_info.addr);
-	__buffer_check(buf, total, len, rest);
-	len = snprintf(buf, rest, "RedirectPort=%s\n", target->redirect_info.port);
-	__buffer_check(buf, total, len, rest);
-	if (target->redirect_info.reason == ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP) {
-		len = snprintf(buf, rest, "RedirectReason=Temporary\n");
-		__buffer_check(buf, total, len, rest);
-	} else if (target->redirect_info.reason == ISCSI_LOGIN_STATUS_TGT_MOVED_PERM) {
-		len = snprintf(buf, rest, "RedirectReason=Permanent\n");
-		__buffer_check(buf, total, len, rest);
-	} else {
-		len = snprintf(buf, rest, "RedirectReason=Unknown\n");
-		__buffer_check(buf, total, len, rest);
-	}
+	concat_printf(b, "RedirectAddress=%s\n", target->redirect_info.addr);
+	concat_printf(b, "RedirectPort=%s\n", target->redirect_info.port);
+	if (target->redirect_info.reason == ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP)
+		concat_printf(b, "RedirectReason=Temporary\n");
+	else if (target->redirect_info.reason == ISCSI_LOGIN_STATUS_TGT_MOVED_PERM)
+		concat_printf(b, "RedirectReason=Permanent\n");
+	else
+		concat_printf(b, "RedirectReason=Unknown\n");
 
-	return total;
+	return adm_err;
 }
 
-static int show_redirect_callback(struct iscsi_target *target, char *buf, int rest)
+static tgtadm_err show_redirect_callback(struct iscsi_target *target, struct concat_buf *b)
 {
-	int len, total = 0;
+	concat_printf(b, "RedirectCallback=%s\n", target->redirect_info.callback);
 
-	len = snprintf(buf, rest, "RedirectCallback=%s\n", target->redirect_info.callback);
-	__buffer_check(buf, total, len, rest);
-
-	return total;
+	return TGTADM_SUCCESS;
 }
 
-int iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid, uint64_t lun,
-		      char *buf, int rest)
+tgtadm_err iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid, uint64_t lun,
+			     struct concat_buf *b)
 {
 	struct iscsi_target* target = NULL;
-	int len, total = 0;
+	tgtadm_err adm_err = TGTADM_INVALID_REQUEST;
 
 	if (mode != MODE_SYSTEM && mode != MODE_PORTAL) {
 	    target = target_find_by_id(tid);
 	    if (!target)
-		    return -TGTADM_NO_TARGET;
+		    return TGTADM_NO_TARGET;
 	}
 
 	switch (mode) {
 	case MODE_SYSTEM:
-		total = isns_show(buf, rest);
+		adm_err = isns_show(b);
 		break;
 	case MODE_TARGET:
 		if (target->redirect_info.callback)
-			len = show_redirect_callback(target, buf, rest);
+			adm_err = show_redirect_callback(target, b);
 		else if (strlen(target->redirect_info.addr))
-			len = show_redirect_info(target, buf, rest);
+			adm_err = show_redirect_info(target, b);
 		else
-			len = show_iscsi_param(buf, target->session_param, rest);
-		total += len;
+			adm_err = show_iscsi_param(target->session_param, b);
 		break;
 	case MODE_SESSION:
-		len = iscsi_target_show_session(target, sid, buf, rest);
-		total += len;
+		adm_err = iscsi_target_show_session(target, sid, b);
 		break;
 	case MODE_PORTAL:
-		len = iscsi_target_show_portals(target, sid, buf, rest);
-		total += len;
+		adm_err = iscsi_target_show_portals(target, sid, b);
 		break;
 	case MODE_CONNECTION:
-		len = iscsi_target_show_connections(target, sid, buf, rest);
-		total += len;
-		break;
-	case MODE_STATS:
-		len = iscsi_target_show_stats(target, sid, buf, rest);
-		total += len;
+		adm_err = iscsi_target_show_connections(target, sid, b);
 		break;
 	default:
 		break;
 	}
 
-	return total;
+	return adm_err;
 }
+
+static void _stat_iscsi_conn_hdr(struct concat_buf *b)
+{
+	concat_printf(b,
+		"sid cid rxdata_octets txdata_octets dataout_pdus datain_pdus cmd_pdus rsp_pdus\n");
+}
+
+static void _stat_iscsi_conn(struct iscsi_connection *conn, struct concat_buf *b)
+{
+	concat_printf(b, "%3d %3d"
+		      " %13" PRIu64
+		      " %13" PRIu64
+		      " %12" PRIu32
+		      " %11" PRIu32
+		      " %8" PRIu32
+		      " %8" PRIu32 "\n",
+		      (unsigned int)conn->session->tsih,
+		      (unsigned int)conn->cid,
+		      conn->stats.rxdata_octets,
+		      conn->stats.txdata_octets,
+		      conn->stats.dataout_pdus,
+		      conn->stats.datain_pdus,
+		      conn->stats.scsicmd_pdus,
+		      conn->stats.scsirsp_pdus);
+}
+
+static tgtadm_err _stat_iscsi_session(struct iscsi_session *session,
+				      uint64_t lun, int filter_lun,
+				      struct concat_buf *b)
+{
+	struct iscsi_target *target = session->target;
+	uint64_t itn_id = session->tsih;
+	int tid = target->tid;	/* global target id */
+	struct it_nexus *itn;
+	struct it_nexus_lu_info *itn_lu;
+	struct scsi_lu *lu;
+	struct iscsi_connection *conn;
+
+	dprintf("tsih:%d lun:%" PRIu64 " filter_lun:%d\n",
+		(unsigned int)session->tsih, lun, filter_lun);
+
+	itn = it_nexus_lookup(tid, itn_id);
+	if (!itn) {
+		eprintf("invalid nexus %d %" PRIx64 "\n", tid, itn_id);
+		return TGTADM_NO_SESSION;
+	}
+
+	tgt_stat_header(b);
+	list_for_each_entry(itn_lu, &itn->itn_itl_info_list, itn_itl_info_siblings) {
+		lu = itn_lu->lu;
+		tgt_stat_line(tid, lu->lun, session->tsih, &itn_lu->stat, b);
+	}
+
+	if (!list_empty(&session->conn_list)) {
+		concat_printf(b, "\n");
+		_stat_iscsi_conn_hdr(b);
+	}
+	list_for_each_entry(conn, &session->conn_list, clist) {
+		_stat_iscsi_conn(conn, b);
+	}
+
+	return TGTADM_SUCCESS;
+}
+
+static tgtadm_err iscsi_stat_connection(uint64_t sid, uint32_t cid, struct concat_buf *b)
+{
+	struct iscsi_session *session;
+	struct iscsi_connection *conn;
+
+	dprintf("sid:%" PRIu64 "cid:%" PRIu32 "\n", sid, cid);
+
+	session = session_lookup_by_tsih((uint16_t)sid);
+	if (!session)
+		return TGTADM_NO_SESSION;
+
+	conn = conn_find(session, cid);
+	if (!conn)
+		return TGTADM_NO_CONNECTION;
+
+	_stat_iscsi_conn_hdr(b);
+	_stat_iscsi_conn(conn, b);
+
+	return TGTADM_SUCCESS;
+}
+
+static tgtadm_err iscsi_stat_session_by_sid(uint64_t sid, struct concat_buf *b)
+{
+	struct iscsi_session *session;
+
+	dprintf("sid:%" PRIu64 "\n", sid);
+
+	session = session_lookup_by_tsih((uint16_t)sid);
+	if (session)
+		return _stat_iscsi_session(session, 0, 0, b);
+	else
+		return TGTADM_NO_SESSION;
+}
+
+static tgtadm_err iscsi_stat_device_by_id(uint64_t lun, uint64_t sid, struct concat_buf *b)
+{
+	struct iscsi_session *session;
+
+	dprintf("lun:%" PRIu64 " sid:%" PRIu64 "\n", lun, sid);
+
+	session = session_lookup_by_tsih((uint16_t)sid);
+	if (session)
+		return _stat_iscsi_session(session, lun, 1, b);
+	else
+		return TGTADM_NO_SESSION;
+}
+
+tgtadm_err iscsi_stat(int mode, int tid, uint64_t sid, uint32_t cid, uint64_t lun,
+		      struct concat_buf *b)
+{
+	tgtadm_err adm_err = TGTADM_INVALID_REQUEST;
+
+	dprintf("mode:%d tid:%d sid:%" PRIu64 " cid:%" PRIu32 " lun:%" PRIx64 "\n",
+		mode, tid, sid, cid, lun);
+
+	switch (mode) {
+	case MODE_DEVICE:
+		adm_err = iscsi_stat_device_by_id(lun, sid, b);
+		break;
+	case MODE_SESSION:
+		adm_err = iscsi_stat_session_by_sid(sid, b);
+		break;
+	case MODE_CONNECTION:
+		adm_err = iscsi_stat_connection(sid, cid, b);
+		break;
+	default:
+		break;
+	}
+
+	return adm_err;
+}
+
