@@ -46,10 +46,10 @@
 
 #define BUFSIZE 4096
 
-static char program_name[] = "tgtadm";
+static const char program_name[] = "tgtadm";
 static int debug;
 
-static const char * tgtadm_strerror(int err)
+static const char *tgtadm_strerror(int err)
 {
 	static const struct {
 		enum tgtadm_errno err;
@@ -63,9 +63,15 @@ static const char * tgtadm_strerror(int err)
 		{ TGTADM_NO_LUN, "can't find the logical unit" },
 		{ TGTADM_NO_SESSION, "can't find the session" },
 		{ TGTADM_NO_CONNECTION, "can't find the connection" },
+		{ TGTADM_NO_BINDING, "can't find the binding" },
 		{ TGTADM_TARGET_EXIST, "this target already exists" },
-		{ TGTADM_LUN_EXIST, "this logical unit number already exists" },
-		{ TGTADM_ACL_EXIST, "this access control rule already exists" },
+		{ TGTADM_BINDING_EXIST, "this binding already exists" },
+		{ TGTADM_LUN_EXIST,
+		  "this logical unit number already exists" },
+		{ TGTADM_ACL_EXIST,
+		  "this access control rule already exists" },
+		{ TGTADM_ACL_NOEXIST,
+		  "this access control rule does not exist" },
 		{ TGTADM_USER_EXIST, "this account already exists" },
 		{ TGTADM_NO_USER, "can't find the account" },
 		{ TGTADM_TOO_MANY_USER, "too many accounts" },
@@ -73,10 +79,14 @@ static const char * tgtadm_strerror(int err)
 		{ TGTADM_OUTACCOUNT_EXIST,
 		  "this target already has an outgoing account" },
 		{ TGTADM_TARGET_ACTIVE, "this target is still active" },
-		{ TGTADM_LUN_ACTIVE, "this logical unit is still active" },
+		{ TGTADM_LUN_ACTIVE,
+		  "this logical unit is still active" },
+		{ TGTADM_DRIVER_ACTIVE, "this driver is busy" },
 		{ TGTADM_UNSUPPORTED_OPERATION,
 		  "this operation isn't supported" },
-		{ TGTADM_UNKNOWN_PARAM, "unknown parameter" }
+		{ TGTADM_UNKNOWN_PARAM, "unknown parameter" },
+		{ TGTADM_PREVENT_REMOVAL,
+		  "this device has Prevent Removal set" }
 	};
 	int i;
 
@@ -103,13 +113,15 @@ struct option const long_options[] = {
 	{"backing-store", required_argument, NULL, 'b'},
 	{"bstype", required_argument, NULL, 'E'},
 	{"bsoflags", required_argument, NULL, 'f'},
+	{"blocksize", required_argument, NULL, 'y'},
 	{"targetname", required_argument, NULL, 'T'},
 	{"initiator-address", required_argument, NULL, 'I'},
+	{"initiator-name", required_argument, NULL, 'Q'},
 	{"user", required_argument, NULL, 'u'},
 	{"password", required_argument, NULL, 'p'},
 	{"host", required_argument, NULL, 'H'},
+	{"force", no_argument, NULL, 'F'},
 	{"params", required_argument, NULL, 'P'},
-
 	{"bus", required_argument, NULL, 'B'},
 	{"device-type", required_argument, NULL, 'Y'},
 	{"outgoing", no_argument, NULL, 'O'},
@@ -117,64 +129,70 @@ struct option const long_options[] = {
 	{NULL, 0, NULL, 0},
 };
 
-static char *short_options = "dhVL:o:m:t:s:c:l:n:v:b:E:f:T:I:u:p:H:P:B:Y:O:C:";
+static char *short_options =
+		"dhVL:o:m:t:s:c:l:n:v:b:E:f:y:T:I:Q:u:p:H:F:P:B:Y:O:C:";
 
 static void usage(int status)
 {
-	if (status != 0)
-		fprintf(stderr, "Try `%s --help' for more information.\n", program_name);
-	else {
-		printf("Usage: %s [OPTION]\n", program_name);
-		printf("\
-Linux SCSI Target Framework Administration Utility, version %s\n\
-\n\
-  --lld <driver> --mode target --op new --tid <id> --targetname <name>\n\
-                        add a new target with <id> and <name>. <id> must not be zero.\n\
-  --lld <driver> --mode target --op delete --tid <id>\n\
-                        delete the specific target with <id>. The target must\n\
-                        have no activity.\n\
-  --lld <driver> --mode target --op show\n\
-                        show all the targets.\n\
-  --lld <driver> --mode target --op show --tid <id>\n\
-                        show the specific target's parameters.\n\
-  --lld <driver> --mode target --op update --tid <id> --name <param> --value <value>\n\
-                        change the target parameters of the specific\n\
-                        target with <id>.\n\
-  --lld <driver> --mode target --op bind --tid <id> --initiator-address <src>\n\
-                        enable the target to accept the specific initiators.\n\
-  --lld <driver> --mode target --op unbind --tid <id> --initiator-address <src>\n\
-                        disable the specific permitted initiators.\n\
-  --lld <driver> --mode logicalunit --op new --tid <id> --lun <lun> \\\n\
-                        --backing-store <path> --bstype <type> --bsoflags <options>\n\
-                        add a new logical unit with <lun> to the specific\n\
-                        target with <id>. The logical unit is offered\n\
-                        to the initiators. <path> must be block device files\n\
-                        (including LVM and RAID devices) or regular files.\n\
-                        bstype option is optional.\n\
-                        bsoflags supported options are sync and direct\n\
-                        (sync:direct for both).\n\
-  --lld <driver> --mode logicalunit --op delete --tid <id> --lun <lun>\n\
-                        delete the specific logical unit with <lun> that\n\
-                        the target with <id> has.\n\
-  --lld <driver> --mode account --op new --user <name> --password <pass>\n\
-                        add a new account with <name> and <pass>.\n\
-  --lld <driver> --mode account --op delete --user <name>\n\
-                        delete the specific account having <name>.\n\
-  --lld <driver> --mode account --op bind --tid <id> --user <name> [--outgoing]\n\
-                        add the specific account having <name> to\n\
-                        the specific target with <id>.\n\
-                        <user> could be <IncomingUser> or <OutgoingUser>.\n\
-                        If you use --outgoing option, the account will\n\
-                        be added as an outgoing account.\n\
-  --lld <driver> --mode account --op unbind --tid <id> --user <name>\n\
-                        delete the specific account having <name> from specific\n\
-                        target.\n\
-  --control-port <port> use control port <port>\n\
-  --help                display this help and exit\n\
-\n\
-Report bugs to <stgt@vger.kernel.org>.\n", TGT_VERSION);
+	if (status != 0) {
+		fprintf(stderr, "Try `%s --help' for more information.\n",
+			program_name);
+		exit(EINVAL);
 	}
-	exit(status == 0 ? 0 : EINVAL);
+
+	printf("Linux SCSI Target administration utility, version %s\n\n"
+		"Usage: %s [OPTION]\n"
+		"--lld <driver> --mode target --op new --tid <id> --targetname <name>\n"
+		"\tadd a new target with <id> and <name>. <id> must not be zero.\n"
+		"--lld <driver> --mode target --op delete [--force] --tid <id>\n"
+		"\tdelete the specific target with <id>.\n"
+		"\tWith force option, the specific target is deleted\n"
+		"\teven if there is an activity.\n"
+		"--lld <driver> --mode target --op show\n"
+		"\tshow all the targets.\n"
+		"--lld <driver> --mode target --op show --tid <id>\n"
+		"\tshow the specific target's parameters.\n"
+		"--lld <driver> --mode target --op update --tid <id> --name <param> --value <value>\n"
+		"\tchange the target parameters of the target with <id>.\n"
+		"--lld <driver> --mode target --op bind --tid <id> --initiator-address <address>\n"
+		"--lld <driver> --mode target --op bind --tid <id> --initiator-name <name>\n"
+		"\tenable the target to accept the specific initiators.\n"
+		"--lld <driver> --mode target --op unbind --tid <id> --initiator-address <address>\n"
+		"--lld <driver> --mode target --op unbind --tid <id> --initiator-name <name>\n"
+		"\tdisable the specific permitted initiators.\n"
+		"--lld <driver> --mode logicalunit --op new --tid <id> --lun <lun>\n"
+		"  --backing-store <path> --bstype <type> --bsoflags <options>\n"
+		"\tadd a new logical unit with <lun> to the specific\n"
+		"\ttarget with <id>. The logical unit is offered\n"
+		"\tto the initiators. <path> must be block device files\n"
+		"\t(including LVM and RAID devices) or regular files.\n"
+		"\tbstype option is optional.\n"
+		"\tbsoflags supported options are sync and direct\n"
+		"\t(sync:direct for both).\n"
+		"--lld <driver> --mode logicalunit --op delete --tid <id> --lun <lun>\n"
+		"\tdelete the specific logical unit with <lun> that\n"
+		"\tthe target with <id> has.\n"
+		"--lld <driver> --mode account --op new --user <name> --password <pass>\n"
+		"\tadd a new account with <name> and <pass>.\n"
+		"--lld <driver> --mode account --op delete --user <name>\n"
+		"\tdelete the specific account having <name>.\n"
+		"--lld <driver> --mode account --op bind --tid <id> --user <name> [--outgoing]\n"
+		"\tadd the specific account having <name> to\n"
+		"\tthe specific target with <id>.\n"
+		"\t<user> could be <IncomingUser> or <OutgoingUser>.\n"
+		"\tIf you use --outgoing option, the account will\n"
+		"\tbe added as an outgoing account.\n"
+		"--lld <driver> --mode account --op unbind --tid <id> --user <name>\n"
+		"\tdelete the specific account having <name> from specific\n"
+		"\ttarget.\n"
+		"--lld <driver> --mode lld --op start\n"
+		"\tStart the specified lld without restarting the tgtd process.\n"
+		"--control-port <port> use control port <port>\n"
+		"--help\n"
+		"\tdisplay this help and exit\n\n"
+		"Report bugs to <stgt@vger.kernel.org>.\n",
+		TGT_VERSION, program_name);
+	exit(0);
 }
 
 static void version(void)
@@ -184,7 +202,7 @@ static void version(void)
 }
 
 /* default port to use for the mgmt channel */
-static short int control_port = 0;
+static short int control_port;
 
 static int ipc_mgmt_connect(int *fd)
 {
@@ -277,29 +295,44 @@ retry:
 	return 0;
 }
 
-static int ipc_mgmt_req(struct tgtadm_req *req)
+static int ipc_mgmt_req(struct tgtadm_req *req, struct concat_buf *b)
 {
-	int err, fd = 0;
+	int err, fd = 0, done = 0;
+
+	req->len = sizeof(*req) + b->size;
 
 	err = ipc_mgmt_connect(&fd);
 	if (err < 0) {
-		eprintf("can't connect to the tgt daemon, %m\n");
+		eprintf("can't connect to tgt daemon, %m\n");
 		goto out;
 	}
 
-	err = write(fd, (char *) req, req->len);
-	if (err < 0) {
-		eprintf("can't send the request to the tgt daemon, %m\n");
+	err = write(fd, req, sizeof(*req));
+	if (err < 0 || err != sizeof(*req)) {
+		eprintf("failed to send request hdr to tgt daemon, %m\n");
 		err = errno;
 		goto out;
 	}
 
-	dprintf("sent to tgtd %d\n", err);
+	while (done < b->size) {
+		err = concat_write(b, fd, done);
+		if (err > 0)
+			done += err;
+		else if (errno != EAGAIN) {
+			eprintf("failed to send request buf to "
+				"tgt daemon, %m\n");
+			err = errno;
+			goto out;
+		}
+	}
+
+	dprintf("sent to tgtd %d\n", req->len);
 
 	err = ipc_mgmt_rsp(fd, req);
 out:
 	if (fd > 0)
 		close(fd);
+	concat_buf_release(b);
 	return err;
 }
 
@@ -381,8 +414,8 @@ static int str_to_mode(char *str)
 		return MODE_CONNECTION;
 	else if (!strcmp("account", str))
 		return MODE_ACCOUNT;
-	else if (!strcmp("stats", str))
-		return MODE_STATS;
+	else if (!strcmp("lld", str))
+		return MODE_LLD;
 	else {
 		eprintf("unknown mode: %s\n", str);
 		exit(1);
@@ -403,12 +436,29 @@ static int str_to_op(char *str)
 		return OP_SHOW;
 	else if (!strcmp("update", str))
 		return OP_UPDATE;
+	else if (!strcmp("stat", str))
+		return OP_STATS;
+	else if (!strcmp("start", str))
+		return OP_START;
+	else if (!strcmp("stop", str))
+		return OP_STOP;
 	else {
 		eprintf("unknown operation: %s\n", str);
 		exit(1);
 	}
 }
 
+static void bad_optarg(int ret, int ch, char *optarg)
+{
+	if (ret == ERANGE)
+		fprintf(stderr, "-%c argument value '%s' out of range\n",
+			ch, optarg);
+	else
+		fprintf(stderr, "-%c argument value '%s' invalid\n",
+			ch, optarg);
+	usage(ret);
+}
+
 static int verify_mode_params(int argc, char **argv, char *allowed)
 {
 	int ch, longindex;
@@ -430,37 +480,28 @@ static int verify_mode_params(int argc, char **argv, char *allowed)
 int main(int argc, char **argv)
 {
 	int ch, longindex, rc;
-	int op, total, tid, rest, mode, dev_type, ac_dir;
+	int op, tid, mode, dev_type, ac_dir;
 	uint32_t cid, hostno;
-	uint64_t sid, lun;
-	char *name, *value, *path, *targetname, *params, *address, *targetOps;
+	uint64_t sid, lun, force;
+	char *name, *value, *path, *targetname, *address, *iqnname, *targetOps;
 	char *portalOps, *bstype;
 	char *bsoflags;
+	char *blocksize;
 	char *user, *password;
-	char *buf;
-	size_t bufsz = BUFSIZE + sizeof(struct tgtadm_req);
-	struct tgtadm_req *req;
+	struct tgtadm_req adm_req = {0}, *req = &adm_req;
+	struct concat_buf b;
 
 	op = tid = mode = -1;
-	total = cid = hostno = sid = 0;
+	cid = hostno = sid = 0;
 	lun = UINT64_MAX;
 
 	rc = 0;
 	dev_type = TYPE_DISK;
 	ac_dir = ACCOUNT_TYPE_INCOMING;
-	rest = BUFSIZE;
-	name = value = path = targetname = address = NULL;
+	name = value = path = targetname = address = iqnname = NULL;
 	targetOps = portalOps = bstype = NULL;
-	bsoflags = user = password = NULL;
-
-	buf = valloc(bufsz);
-	if (!buf) {
-		eprintf("%s\n",	tgtadm_strerror(TGTADM_NOMEM));
-		return ENOMEM;
-	}
-
-	memset(buf, 0, bufsz);
-	req = (struct tgtadm_req *) buf;
+	bsoflags = blocksize = user = password = NULL;
+	force = 0;
 
 	optind = 1;
 	while ((ch = getopt_long(argc, argv, short_options,
@@ -477,20 +518,24 @@ int main(int argc, char **argv)
 			mode = str_to_mode(optarg);
 			break;
 		case 't':
-			tid = strtol(optarg, NULL, 10);
-			if (errno == EINVAL) {
-				eprintf("invalid tid '%s'\n", optarg);
-				exit(EINVAL);
-			}
+			rc = str_to_int_ge(optarg, tid, 0);
+			if (rc)
+				bad_optarg(rc, ch, optarg);
 			break;
 		case 's':
-			sid = strtoull(optarg, NULL, 10);
+			rc = str_to_int(optarg, sid);
+			if (rc)
+				bad_optarg(rc, ch, optarg);
 			break;
 		case 'c':
-			cid = strtoul(optarg, NULL, 10);
+			rc = str_to_int(optarg, cid);
+			if (rc)
+				bad_optarg(rc, ch, optarg);
 			break;
 		case 'l':
-			lun = strtoull(optarg, NULL, 10);
+			rc = str_to_int(optarg, lun);
+			if (rc)
+				bad_optarg(rc, ch, optarg);
 			break;
 		case 'P':
 			if (mode == MODE_PORTAL)
@@ -513,6 +558,9 @@ int main(int argc, char **argv)
 		case 'I':
 			address = optarg;
 			break;
+		case 'Q':
+			iqnname = optarg;
+			break;
 		case 'u':
 			user = optarg;
 			break;
@@ -523,11 +571,19 @@ int main(int argc, char **argv)
 			hostno = bus_to_host(optarg);
 			break;
 		case 'H':
-			hostno = strtol(optarg, NULL, 10);
+			rc = str_to_int_ge(optarg, hostno, 0);
+			if (rc)
+				bad_optarg(rc, ch, optarg);
+			break;
+		case 'F':
+			force = 1;
 			break;
 		case 'f':
 			bsoflags = optarg;
 			break;
+		case 'y':
+			blocksize = optarg;
+			break;
 		case 'E':
 			bstype = optarg;
 			break;
@@ -538,7 +594,9 @@ int main(int argc, char **argv)
 			ac_dir = ACCOUNT_TYPE_OUTGOING;
 			break;
 		case 'C':
-			control_port = strtol(optarg, NULL, 10);
+			rc = str_to_int_gt(optarg, control_port, 0);
+			if (rc)
+				bad_optarg(rc, ch, optarg);
 			break;
 		case 'V':
 			version();
@@ -586,6 +644,7 @@ int main(int argc, char **argv)
 			break;
 		case OP_SHOW:
 		case OP_DELETE:
+		case OP_STATS:
 			break;
 		default:
 			eprintf("option %d not supported in system mode\n", op);
@@ -596,7 +655,11 @@ int main(int argc, char **argv)
 
 	if (mode == MODE_TARGET) {
 		if ((tid <= 0 && (op != OP_SHOW))) {
-			eprintf("'tid' option is necessary\n");
+			if (tid == 0)
+				eprintf("'tid' cannot be 0\n");
+			else
+				eprintf("'tid' option is necessary\n");
+
 			exit(EINVAL);
 		}
 		switch (op) {
@@ -604,34 +667,43 @@ int main(int argc, char **argv)
 			rc = verify_mode_params(argc, argv, "LmotTC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
-					  "allowed/supported\n", rc);
+					"allowed/supported\n", rc);
 				exit(EINVAL);
 			}
 			if (!targetname) {
-				eprintf("creating a new target requires name\n");
+				eprintf("creating new target requires "
+					"a name, use --targetname\n");
 				exit(EINVAL);
 			}
 			break;
 		case OP_DELETE:
+			rc = verify_mode_params(argc, argv, "LmotCF");
+			if (rc) {
+				eprintf("target mode: option '-%c' is not "
+					"allowed/supported\n", rc);
+				exit(EINVAL);
+			}
+			break;
 		case OP_SHOW:
+		case OP_STATS:
 			rc = verify_mode_params(argc, argv, "LmotC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
-					  "allowed/supported\n", rc);
+					"allowed/supported\n", rc);
 				exit(EINVAL);
 			}
 			break;
 		case OP_BIND:
 		case OP_UNBIND:
-			rc = verify_mode_params(argc, argv, "LmotIBHC");
+			rc = verify_mode_params(argc, argv, "LmotIQBHC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
 					  "allowed/supported\n", rc);
 				exit(EINVAL);
 			}
-			if (!address && !hostno) {
+			if (!address && !iqnname && !hostno) {
 				eprintf("%s operation requires"
-					" initiator-address or bus\n",
+					" initiator-address, initiator-name or bus\n",
 					op == OP_BIND ? "bind" : "unbind");
 				exit(EINVAL);
 			}
@@ -639,12 +711,12 @@ int main(int argc, char **argv)
 		case OP_UPDATE:
 			rc = verify_mode_params(argc, argv, "LmotnvC");
 			if (rc) {
-				eprintf("target mode: option '-%c' is not "
+				eprintf("target mode: option '-%c' is not " \
 					  "allowed/supported\n", rc);
 				exit(EINVAL);
 			}
 			if ((!name || !value)) {
-				eprintf("update operation requires 'name'"
+				eprintf("update operation requires 'name'" \
 						" and 'value' options\n");
 				exit(EINVAL);
 			}
@@ -661,12 +733,13 @@ int main(int argc, char **argv)
 		case OP_NEW:
 			rc = verify_mode_params(argc, argv, "LmoupfC");
 			if (rc) {
-				eprintf("logicalunit mode: option '-%c' is not "
-					  "allowed/supported\n", rc);
+				eprintf("logicalunit mode: option '-%c' is "
+					"not allowed/supported\n", rc);
 				exit(EINVAL);
 			}
 			if (!user || !password) {
-				eprintf("'user' and 'password' options is necessary\n");
+				eprintf("'user' and 'password' options "
+					"are required\n");
 				exit(EINVAL);
 			}
 			break;
@@ -674,7 +747,7 @@ int main(int argc, char **argv)
 			rc = verify_mode_params(argc, argv, "LmoC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
-					  "allowed/supported\n", rc);
+					"allowed/supported\n", rc);
 				exit(EINVAL);
 			}
 			break;
@@ -682,7 +755,7 @@ int main(int argc, char **argv)
 			rc = verify_mode_params(argc, argv, "LmouC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
-					  "allowed/supported\n", rc);
+					"allowed/supported\n", rc);
 				exit(EINVAL);
 			}
 			break;
@@ -690,7 +763,7 @@ int main(int argc, char **argv)
 			rc = verify_mode_params(argc, argv, "LmotuOC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
-					  "allowed/supported\n", rc);
+					"allowed/supported\n", rc);
 				exit(EINVAL);
 			}
 			if (!user) {
@@ -701,10 +774,10 @@ int main(int argc, char **argv)
 				tid = GLOBAL_TID;
 			break;
 		case OP_UNBIND:
-			rc = verify_mode_params(argc, argv, "LmotuC");
+			rc = verify_mode_params(argc, argv, "LmotuOC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
-					  "allowed/supported\n", rc);
+					"allowed/supported\n", rc);
 				exit(EINVAL);
 			}
 			if (!user) {
@@ -715,7 +788,8 @@ int main(int argc, char **argv)
 				tid = GLOBAL_TID;
 			break;
 		default:
-			eprintf("option %d not supported in account mode\n", op);
+			eprintf("option %d not supported in account mode\n",
+				op);
 			exit(EINVAL);
 			break;
 		}
@@ -723,7 +797,10 @@ int main(int argc, char **argv)
 
 	if (mode == MODE_DEVICE) {
 		if (tid <= 0) {
-			eprintf("'tid' option is necessary\n");
+			if (tid == 0)
+				eprintf("'tid' must not be 0\n");
+			else
+				eprintf("'tid' option is necessary\n");
 			exit(EINVAL);
 		}
 		if (lun == UINT64_MAX) {
@@ -732,19 +809,22 @@ int main(int argc, char **argv)
 		}
 		switch (op) {
 		case OP_NEW:
-			rc = verify_mode_params(argc, argv, "LmoftlbEYC");
+			rc = verify_mode_params(argc, argv, "LmofytlbEYC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
 					  "allowed/supported\n", rc);
 				exit(EINVAL);
 			}
-			if (!path) {
+			if (!path && dev_type != TYPE_MMC
+			    && dev_type != TYPE_TAPE
+			    && dev_type != TYPE_DISK) {
 				eprintf("'backing-store' option "
 						"is necessary\n");
 				exit(EINVAL);
 			}
 			break;
 		case OP_DELETE:
+		case OP_STATS:
 			rc = verify_mode_params(argc, argv, "LmotlC");
 			if (rc) {
 				eprintf("target mode: option '-%c' is not "
@@ -753,7 +833,7 @@ int main(int argc, char **argv)
 			}
 			break;
 		case OP_UPDATE:
-			rc = verify_mode_params(argc, argv, "LmoftlPC");
+			rc = verify_mode_params(argc, argv, "LmofytlPC");
 			if (rc) {
 				eprintf("option '-%c' not supported in "
 					"logicalunit mode\n", rc);
@@ -762,7 +842,7 @@ int main(int argc, char **argv)
 			break;
 		default:
 			eprintf("option %d not supported in "
-					"logicalunit mode\n", op);
+				"logicalunit mode\n", op);
 			exit(EINVAL);
 			break;
 		}
@@ -812,6 +892,25 @@ int main(int argc, char **argv)
 		}
 	}
 
+	if (mode == MODE_LLD) {
+		switch (op) {
+		case OP_START:
+		case OP_STOP:
+		case OP_SHOW:
+			rc = verify_mode_params(argc, argv, "LmoC");
+			if (rc) {
+				eprintf("system mode: option '-%c' is not "
+					"allowed/supported\n", rc);
+				exit(EINVAL);
+			}
+			break;
+		default:
+			eprintf("option %d not supported in lld mode\n", op);
+			exit(EINVAL);
+			break;
+		}
+	}
+
 	req->op = op;
 	req->tid = tid;
 	req->sid = sid;
@@ -821,48 +920,60 @@ int main(int argc, char **argv)
 	req->host_no = hostno;
 	req->device_type = dev_type;
 	req->ac_dir = ac_dir;
+	req->force = force;
 
-	params = buf + sizeof(*req);
+	concat_buf_init(&b);
 
 	if (name)
-		shprintf(total, params, rest, "%s=%s", name, value);
+		concat_printf(&b, "%s=%s", name, value);
 	if (path)
-		shprintf(total, params, rest, "%spath=%s",
-			 rest == BUFSIZE ? "" : ",", path);
+		concat_printf(&b, "%spath=%s", concat_delim(&b, ","), path);
 
 	if (req->device_type == TYPE_TAPE)
-		shprintf(total, params, rest, "%sbstype=%s",
-			 rest == BUFSIZE ? "" : ",", "ssc");
+		concat_printf(&b, "%sbstype=%s", concat_delim(&b, ","),
+			      "ssc");
 	else if (bstype)
-		shprintf(total, params, rest, "%sbstype=%s",
-			 rest == BUFSIZE ? "" : ",", bstype);
+		concat_printf(&b, "%sbstype=%s", concat_delim(&b, ","),
+			      bstype);
 	if (bsoflags)
-		shprintf(total, params, rest, "%sbsoflags=%s",
-			 rest == BUFSIZE ? "" : ",", bsoflags);
+		concat_printf(&b, "%sbsoflags=%s", concat_delim(&b, ","),
+			      bsoflags);
+	if (blocksize)
+		concat_printf(&b, "%sblocksize=%s", concat_delim(&b, ","),
+			      blocksize);
 	if (targetname)
-		shprintf(total, params, rest, "%stargetname=%s",
-			 rest == BUFSIZE ? "" : ",", targetname);
+		concat_printf(&b, "%stargetname=%s", concat_delim(&b, ","),
+			      targetname);
 	if (address)
-		shprintf(total, params, rest, "%sinitiator-address=%s",
-			 rest == BUFSIZE ? "" : ",", address);
+		concat_printf(&b, "%sinitiator-address=%s",
+			      concat_delim(&b, ","), address);
+	if (iqnname)
+		concat_printf(&b, "%sinitiator-name=%s", concat_delim(&b, ","),
+			      iqnname);
 	if (user)
-		shprintf(total, params, rest, "%suser=%s",
-			 rest == BUFSIZE ? "" : ",", user);
+		concat_printf(&b, "%suser=%s", concat_delim(&b, ","),
+			      user);
 	if (password)
-		shprintf(total, params, rest, "%spassword=%s",
-			 rest == BUFSIZE ? "" : ",", password);
+		concat_printf(&b, "%spassword=%s", concat_delim(&b, ","),
+			      password);
 	/* Trailing ',' makes parsing params in modules easier.. */
 	if (targetOps)
-		shprintf(total, params, rest, "%stargetOps %s,",
-			 rest == BUFSIZE ? "" : ",", targetOps);
+		concat_printf(&b, "%stargetOps %s,", concat_delim(&b, ","),
+			      targetOps);
 	if (portalOps)
-		shprintf(total, params, rest, "%sportalOps %s,",
-			 rest == BUFSIZE ? "" : ",", portalOps);
+		concat_printf(&b, "%sportalOps %s,", concat_delim(&b, ","),
+			      portalOps);
 
-	req->len = sizeof(*req) + total;
+	if (b.err) {
+		eprintf("BUFSIZE (%d bytes) isn't long enough\n", BUFSIZE);
+		return EINVAL;
+	}
 
-	return ipc_mgmt_req(req);
-overflow:
-	eprintf("BUFSIZE (%d bytes) isn't long enough\n", BUFSIZE);
-	return EINVAL;
+	rc = concat_buf_finish(&b);
+	if (rc) {
+		eprintf("failed to create request, errno:%d\n", rc);
+		exit(rc);
+	}
+
+	return ipc_mgmt_req(req, &b);
 }
