@@ -35,6 +35,7 @@
 #include <sys/epoll.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <sys/stat.h>
 
 #include "list.h"
 #include "tgtd.h"
@@ -50,34 +51,48 @@ static char program_name[] = "tgtd";
 static LIST_HEAD(tgt_events_list);
 static LIST_HEAD(tgt_sched_events_list);
 
-static struct option const long_options[] =
-{
+static struct option const long_options[] = {
 	{"foreground", no_argument, 0, 'f'},
 	{"control-port", required_argument, 0, 'C'},
+	{"nr_iothreads", required_argument, 0, 't'},
 	{"debug", required_argument, 0, 'd'},
 	{"version", no_argument, 0, 'V'},
 	{"help", no_argument, 0, 'h'},
 	{0, 0, 0, 0},
 };
 
-static char *short_options = "fC:d:Vh";
+static char *short_options = "fC:d:t:Vh";
+static char *spare_args;
 
 static void usage(int status)
 {
-	if (status)
-		fprintf(stderr, "Try `%s --help' for more information.\n", program_name);
-	else {
-		printf("Usage: %s [OPTION]\n", program_name);
-		printf("\
-Target framework daemon, version %s\n\
-  -f, --foreground        make the program run in the foreground\n\
-  -C, --control-port NNNN use port NNNN for the mgmt channel\n\
-  -d, --debug debuglevel  print debugging information\n\
-  -V, --version           print version and exit\n\
-  -h, --help              display this help and exit\n\
-", TGT_VERSION);
+	if (status) {
+		fprintf(stderr, "Try `%s --help' for more information.\n",
+			program_name);
+		exit(status);
 	}
-	exit(status);
+
+	printf("Linux SCSI Target framework daemon, version %s\n\n"
+		"Usage: %s [OPTION]\n"
+		"-f, --foreground        make the program run in the foreground\n"
+		"-C, --control-port NNNN use port NNNN for the mgmt channel\n"
+		"-t, --nr_iothreads NNNN specify the number of I/O threads\n"
+		"-d, --debug debuglevel  print debugging information\n"
+		"-V, --version           print version and exit\n"
+		"-h, --help              display this help and exit\n",
+		TGT_VERSION, program_name);
+	exit(0);
+}
+
+static void bad_optarg(int ret, int ch, char *optarg)
+{
+	if (ret == ERANGE)
+		fprintf(stderr, "-%c argument value '%s' out of range\n",
+			ch, optarg);
+	else
+		fprintf(stderr, "-%c argument value '%s' invalid\n",
+			ch, optarg);
+	usage(ret);
 }
 
 static void version(void)
@@ -87,26 +102,39 @@ static void version(void)
 }
 
 /* Default TGT mgmt port */
-short int control_port = 0;
+short int control_port;
 
-static void signal_catch(int signo) {
+static void signal_catch(int signo)
+{
 }
 
 static int oom_adjust(void)
 {
 	int fd, err;
-	char path[64];
+	const char *path, *score;
+	struct stat st;
 
 	/* Avoid oom-killer */
-	sprintf(path, "/proc/%d/oom_adj", getpid());
+	path = "/proc/self/oom_score_adj";
+	score = "-1000\n";
+
+	if (stat(path, &st)) {
+		/* oom_score_adj cannot be used, try oom_adj */
+		path = "/proc/self/oom_adj";
+		score = "-17\n";
+	}
+
 	fd = open(path, O_WRONLY);
 	if (fd < 0) {
-		fprintf(stderr, "can't adjust oom-killer's pardon %s, %m\n", path);
+		fprintf(stderr, "can't adjust oom-killer's pardon %s, %m\n",
+			path);
 		return errno;
 	}
-	err = write(fd, "-17\n", 4);
+
+	err = write(fd, score, strlen(score));
 	if (err < 0) {
-		fprintf(stderr, "can't adjust oom-killer's pardon %s, %m\n", path);
+		fprintf(stderr, "can't adjust oom-killer's pardon %s, %m\n",
+			path);
 		close(fd);
 		return errno;
 	}
@@ -405,22 +433,30 @@ retry:
 		goto retry;
 }
 
-static int lld_init(int *use_kernel, char *args)
+int lld_init_one(int lld_index)
 {
-	int i, err, nr;
+	int err;
+
+	INIT_LIST_HEAD(&tgt_drivers[lld_index]->target_list);
+	if (tgt_drivers[lld_index]->init) {
+		err = tgt_drivers[lld_index]->init(lld_index, spare_args);
+		if (err) {
+			tgt_drivers[lld_index]->drv_state = DRIVER_ERR;
+			return err;
+		}
+		tgt_drivers[lld_index]->drv_state = DRIVER_INIT;
+	}
+	return 0;
+}
+
+static int lld_init(void)
+{
+	int i, nr;
 
 	for (i = nr = 0; tgt_drivers[i]; i++) {
-		if (tgt_drivers[i]->init) {
-			err = tgt_drivers[i]->init(i, args);
-			if (err)
-				continue;
-		}
-
-		if (tgt_drivers[i]->use_kernel)
-			(*use_kernel)++;
-		nr++;
+		if (!lld_init_one(i))
+			nr++;
 	}
-
 	return nr;
 }
 
@@ -431,6 +467,7 @@ static void lld_exit(void)
 	for (i = 0; tgt_drivers[i]; i++) {
 		if (tgt_drivers[i]->exit)
 			tgt_drivers[i]->exit();
+		tgt_drivers[i]->drv_state = DRIVER_EXIT;
 	}
 }
 
@@ -478,14 +515,11 @@ int main(int argc, char **argv)
 	struct sigaction sa_new;
 	int err, ch, longindex, nr_lld = 0;
 	int is_daemon = 1, is_debug = 0;
-	int use_kernel = 0;
 	int ret;
 
-	/* do not allow ctrl-c for now... */
 	sa_new.sa_handler = signal_catch;
 	sigemptyset(&sa_new.sa_mask);
 	sa_new.sa_flags = 0;
-	sigaction(SIGINT, &sa_new, &sa_old);
 	sigaction(SIGPIPE, &sa_new, &sa_old);
 	sigaction(SIGTERM, &sa_new, &sa_old);
 
@@ -503,10 +537,19 @@ int main(int argc, char **argv)
 			is_daemon = 0;
 			break;
 		case 'C':
-			control_port = atoi(optarg);
+			ret = str_to_int_gt(optarg, control_port, 0);
+			if (ret)
+				bad_optarg(ret, ch, optarg);
+			break;
+		case 't':
+			ret = str_to_int_gt(optarg, nr_iothreads, 0);
+			if (ret)
+				bad_optarg(ret, ch, optarg);
 			break;
 		case 'd':
-			is_debug = atoi(optarg);
+			ret = str_to_int_range(optarg, is_debug, 0, 1);
+			if (ret)
+				bad_optarg(ret, ch, optarg);
 			break;
 		case 'V':
 			version();
@@ -532,19 +575,25 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
-	nr_lld = lld_init(&use_kernel, argv[optind]);
-	if (!nr_lld) {
-		fprintf(stderr, "No available low level driver!\n");
-		exit(1);
-	}
+	spare_args = optind < argc ? argv[optind] : NULL;
 
 	err = ipc_init();
 	if (err)
 		exit(1);
 
-	if (is_daemon && daemon(0,0))
+	if (is_daemon && daemon(0, 0))
 		exit(1);
 
+	err = log_init(program_name, LOG_SPACE_SIZE, is_daemon, is_debug);
+	if (err)
+		exit(1);
+
+	nr_lld = lld_init();
+	if (!nr_lld) {
+		fprintf(stderr, "No available low level driver!\n");
+		exit(1);
+	}
+
 	err = oom_adjust();
 	if (err)
 		exit(1);
@@ -553,18 +602,6 @@ int main(int argc, char **argv)
 	if (err)
 		exit(1);
 
-	err = log_init(program_name, LOG_SPACE_SIZE, is_daemon, is_debug);
-	if (err)
-		exit(1);
-
-	if (use_kernel) {
-		err = kreq_init();
-		if (err) {
-			eprintf("No kernel interface\n");
-			exit(1);
-		}
-	}
-
 	err = work_timer_start();
 	if (err)
 		exit(1);
