@@ -5,9 +5,15 @@
 #include <endian.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <linux/falloc.h>
 #include <signal.h>
 #include <syscall.h>
 #include <unistd.h>
+#include <stdarg.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <limits.h>
 #include <linux/types.h>
 
 #include "be_byteshift.h"
@@ -56,10 +62,13 @@
 
 extern int get_blk_shift(unsigned int size);
 extern int chrdev_open(char *modname, char *devpath, uint8_t minor, int *fd);
-extern int backed_file_open(char *path, int oflag, uint64_t *size);
+extern int backed_file_open(char *path, int oflag, uint64_t *size,
+				uint32_t *blksize);
 extern int set_non_blocking(int fd);
 extern int str_to_open_flags(char *buf);
 extern char *open_flags_to_str(char *dest, int flags);
+extern int spc_memcpy(uint8_t *dst, uint32_t *dst_remain_len,
+		      uint8_t *src, uint32_t src_len);
 
 #define zalloc(size)			\
 ({					\
@@ -87,40 +96,8 @@ static inline int between(uint32_t seq1, uint32_t seq2, uint32_t seq3)
 	return seq3 - seq2 >= seq1 - seq2;
 }
 
-#define shprintf(total, buf, rest, fmt, args...)			\
-do {									\
-	int len;							\
-	len = snprintf(buf, rest, fmt, ##args);				\
-	if (len > rest)							\
-		goto overflow;						\
-	buf += len;							\
-	total += len;							\
-	rest -= len;							\
-} while (0)
-
 extern unsigned long pagesize, pageshift;
 
-#if defined (SYNC_FILE_RANGE_WAIT_BEFORE) && \
-	(defined(__NR_sync_file_range) || defined(__NR_sync_file_range2))
-static inline int __sync_file_range(int fd, __off64_t offset, __off64_t bytes)
-{
-	int ret;
-#if defined(__NR_sync_file_range)
-	long int n = __NR_sync_file_range;
-#else
-	long int n = __NR_sync_file_range2;
-#endif
-	unsigned int flags = SYNC_FILE_RANGE_WAIT_BEFORE | SYNC_FILE_RANGE_WRITE
-		| SYNC_FILE_RANGE_WAIT_AFTER;
-	ret = syscall(n, fd, offset, bytes, flags);
-	if (ret)
-		ret = fsync(fd);
-	return ret;
-}
-#else
-#define __sync_file_range(fd, offset, bytes) fsync(fd)
-#endif
-
 #if defined(__NR_signalfd) && defined(USE_SIGNALFD)
 
 /*
@@ -159,4 +136,88 @@ struct signalfd_siginfo {
 };
 #endif
 
+/* convert string to integer, check for validity of the string numeric format
+ * and the natural boundaries of the integer value type (first get a 64-bit
+ * value and check that it fits the range of the destination integer).
+ */
+#define str_to_int(str, val)    			\
+({      						\
+	int ret = 0;    				\
+	char *ptr;      				\
+	unsigned long long ull_val;     		\
+	ull_val = strtoull(str, &ptr, 0);       	\
+	val = (typeof(val)) ull_val;    		\
+	if (errno || ptr == str)			\
+		ret = EINVAL;   			\
+	else if (val != ull_val)			\
+		ret = ERANGE;   			\
+	ret;						\
+})
+
+/* convert to int and check: strictly greater than */
+#define str_to_int_gt(str, val, minv)   		\
+({      						\
+	int ret = str_to_int(str, val); 		\
+	if (!ret && (val < minv))       		\
+		ret = ERANGE;   			\
+	ret;						\
+})
+
+/* convert and check: greater than or equal  */
+#define str_to_int_ge(str, val, minv)   		\
+({      						\
+	int ret = str_to_int(str, val); 		\
+	if (!ret && (val < minv))       		\
+		ret = ERANGE;   			\
+	ret;    					\
+})
+
+/* convert and check: strictly less than  */
+#define str_to_int_lt(str, val, maxv)   		\
+({      						\
+	int ret = str_to_int(str, val); 		\
+	if (!ret && (val > maxv))       		\
+		ret = ERANGE;				\
+	ret;						\
+})
+
+/* convert and check: range, ends inclusive  */
+#define str_to_int_range(str, val, minv, maxv)  	\
+({      						\
+	int ret = str_to_int(str, val); 		\
+	if (!ret && (val < minv || val > maxv)) 	\
+		ret = ERANGE;   			\
+	ret;						\
+})
+
+struct concat_buf {
+	FILE *streamf;
+	int err;
+	int used;
+	char *buf;
+	size_t size;
+};
+
+void concat_buf_init(struct concat_buf *b);
+int concat_printf(struct concat_buf *b, const char *format, ...);
+const char *concat_delim(struct concat_buf *b, const char *delim);
+int concat_buf_finish(struct concat_buf *b);
+int concat_write(struct concat_buf *b, int fd, int offset);
+void concat_buf_release(struct concat_buf *b);
+
+
+/* If we have recent enough glibc to support PUNCH HOLE we try to unmap
+ * the region.
+ */
+static inline int unmap_file_region(int fd, off_t offset, off_t length)
+{
+#ifdef FALLOC_FL_PUNCH_HOLE
+	if (fallocate(fd, FALLOC_FL_KEEP_SIZE | FALLOC_FL_PUNCH_HOLE,
+			offset, length) == 0)
+		return 0; 
+#endif
+	return -1;
+}
+
+
 #endif
